from gettext import gettext as _

from gi.repository import Gtk

from lutris.gui.installer.widgets import InstallerLabel
from lutris.util.strings import gtk_safe, gtk_safe_urls


class InstallerScriptBox(Gtk.VBox):
    """Box displaying the details of a script, with associated action buttons"""

    def __init__(self, script, parent=None, revealed=False):
        super().__init__()
        self.script = script
        self.parent = parent
        self.revealer = None
        self.set_margin_left(12)
        self.set_margin_right(12)
        box = Gtk.Box(spacing=12, margin_top=6, margin_bottom=6)
        box.pack_start(self.get_infobox(), True, True, 0)
        box.add(self.get_install_button())
        self.add(box)
        self.add(self.get_revealer(revealed))

    def get_rating(self):
        """Return a string representation of the API rating"""
        return ""

    def get_infobox(self):
        """Return the central information box"""
        info_box = Gtk.VBox(spacing=6)
        title_box = Gtk.HBox(spacing=6)
        runner_label = InstallerLabel("%s" % self.script["runner"])
        runner_label.get_style_context().add_class("info-pill")
        title_box.pack_start(runner_label, False, False, 0)
        title_box.add(InstallerLabel("<b>%s</b>" % gtk_safe(self.script["version"]), selectable=True))
        title_box.pack_start(InstallerLabel(""), True, True, 0)
        rating_label = InstallerLabel(self.get_rating(), selectable=True)
        rating_label.set_alignment(1, 0.5)
        title_box.pack_end(rating_label, False, False, 0)
        info_box.add(title_box)
        info_box.add(self.get_credits())
        info_box.add(InstallerLabel(gtk_safe_urls(self.script["description"]), selectable=True))

        return info_box

    def get_revealer(self, revealed):
        """Return the revelaer widget"""
        self.revealer = Gtk.Revealer()
        box = Gtk.VBox(visible=True)
        box.add(self.get_notes())

        self.revealer.add(box)
        self.revealer.set_reveal_child(revealed)
        return self.revealer

    def get_install_button(self):
        """Return the install button widget"""
        align = Gtk.Alignment()
        align.set(0, 0, 0, 0)

        install_button = Gtk.Button(_("Install"))
        install_button.connect("clicked", self.on_install_clicked)
        style_context = install_button.get_style_context()
        style_context.add_class("suggested-action")
        align.add(install_button)
        return align

    def get_notes(self):
        """Return the notes widget"""
        notes = self.script["notes"].strip()
        if not notes:
            return Gtk.Alignment()
        return self._get_installer_label(notes)

    def get_credits(self):
        credits_text = self.script.get("credits", "").strip()
        if not credits_text:
            return Gtk.Alignment()
        return self._get_installer_label(gtk_safe_urls(credits_text))

    def _get_installer_label(self, text):
        _label = InstallerLabel(text, selectable=True)
        _label.set_margin_top(12)
        _label.set_margin_bottom(12)
        _label.set_margin_right(12)
        _label.set_margin_left(12)
        return _label

    def reveal(self, reveal=True):
        """Show or hide the information in the revealer"""
        if self.revealer:
            self.revealer.set_reveal_child(reveal)

    def on_install_clicked(self, _widget):
        """Handler to notify the parent of the selected installer"""
        self.parent.emit("installer-selected", self.script["version"])



# Standard Library
import os
import shutil
from gettext import gettext as _

# Lutris Modules
from lutris.config import LutrisConfig
from lutris.exceptions import MissingBiosError, MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import system


class hatari(Runner):
    human_name = _("Hatari")
    description = _("Atari ST computers emulator")
    platforms = [_("Atari ST")]
    runnable_alone = True
    flatpak_id = "org.tuxfamily.hatari"
    runner_executable = "hatari/bin/hatari"
    entry_point_option = "disk-a"

    game_options = [
        {
            "option": "disk-a",
            "type": "file",
            "label": _("Floppy Disk A"),
            "help": _(
                "Hatari supports floppy disk images in the following "
                "formats: ST, DIM, MSA, STX, IPF, RAW and CRT. The last "
                "three require the caps library (capslib). ZIP is "
                "supported, you don't need to uncompress the file."
            ),
        },
        {
            "option": "disk-b",
            "type": "file",
            "label": _("Floppy Disk B"),
            "help": _(
                "Hatari supports floppy disk images in the following "
                "formats: ST, DIM, MSA, STX, IPF, RAW and CRT. The last "
                "three require the caps library (capslib). ZIP is "
                "supported, you don't need to uncompress the file."
            ),
        },
    ]

    joystick_choices = [(_("None"), "none"), (_("Keyboard"), "keys"), (_("Joystick"), "real")]

    runner_options = [
        {
            "option": "bios_file",
            "type": "file",
            "label": _("Bios file (TOS)"),
            "help": _(
                "TOS is the operating system of the Atari ST "
                "and is necessary to run applications with the best "
                "fidelity, minimizing risks of issues.\n"
                "TOS 1.02 is recommended for games."
            ),
        },
        {
            "option": "fullscreen",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Fullscreen"),
            "default": False,
        },
        {
            "option": "zoom",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Scale up display by 2 (Atari ST/STE)"),
            "default": True,
            "help": _("Double the screen size in windowed mode."),
        },
        {
            "option": "borders",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Add borders to display"),
            "default": False,
            "help": _(
                "Useful for some games and demos using the overscan "
                "technique. The Atari ST displayed borders around the "
                "screen because it was not powerful enough to display "
                "graphics in fullscreen. But people from the demo scene "
                "were able to remove them and some games made use of "
                "this technique."
            ),
        },
        {
            "option": "status",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Display status bar"),
            "default": False,
            "help": _(
                "Displays a status bar with some useful information, "
                "like green leds lighting up when the floppy disks are "
                "read."
            ),
        },
        {
            "option": "joy0",
            "type": "choice",
            "section": _("Joysticks"),
            "label": _("Joystick 0"),
            "choices": joystick_choices,
            "default": "none",
        },
        {
            "option": "joy1",
            "type": "choice",
            "section": _("Joysticks"),
            "label": _("Joystick 1"),
            "choices": joystick_choices,
            "default": "real",
        },
    ]

    def install(self, install_ui_delegate, version=None, callback=None):
        def on_runner_installed(*args):
            bios_path = system.create_folder("~/.hatari/bios")

            bios_filename = install_ui_delegate.show_install_file_inquiry(
                question=_("Do you want to select an Atari ST BIOS file?"),
                title=_("Use BIOS file?"),
                message=_("Select a BIOS file"),
            )

            if bios_filename:
                shutil.copy(bios_filename, bios_path)
                bios_path = os.path.join(bios_path, os.path.basename(bios_filename))
                config = LutrisConfig(runner_slug="hatari")
                config.raw_runner_config.update({"bios_file": bios_path})
                config.save()
            if callback:
                callback()

        super().install(install_ui_delegate, version=version, callback=on_runner_installed)

    def play(self):  # pylint: disable=too-many-branches
        params = self.get_command()
        if self.runner_config.get("fullscreen"):
            params.append("--fullscreen")
        else:
            params.append("--window")

        params.append("--zoom")
        if self.runner_config.get("zoom"):
            params.append("2")
        else:
            params.append("1")

        params.append("--borders")
        if self.runner_config.get("borders"):
            params.append("true")
        else:
            params.append("false")

        params.append("--statusbar")
        if self.runner_config.get("status"):
            params.append("true")
        else:
            params.append("false")

        if self.runner_config.get("joy0"):
            params.append("--joy0")
            params.append(self.runner_config["joy0"])

        if self.runner_config.get("joy1"):
            params.append("--joy1")
            params.append(self.runner_config["joy1"])

        if system.path_exists(self.runner_config.get("bios_file", "")):
            params.append("--tos")
            params.append(self.runner_config["bios_file"])
        else:
            raise MissingBiosError()
        diska = self.game_config.get("disk-a")
        if not system.path_exists(diska):
            raise MissingGameExecutableError(filename=diska)
        params.append("--disk-a")
        params.append(diska)

        diskb = self.game_config.get("disk-b")
        if diskb:
            params.append("--disk-b")
            params.append(diskb)

        return {"command": params}



"""XDG applications service"""

import json
import os
import re
import shlex
import subprocess
from gettext import gettext as _

from gi.repository import Gio

from lutris import settings
from lutris.database.games import get_games_where
from lutris.services.base import BaseService
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util import system
from lutris.util.log import logger
from lutris.util.strings import slugify


def get_appid(app):
    """Get the appid for the game"""
    try:
        return os.path.splitext(app.get_id())[0]
    except UnicodeDecodeError:
        logger.exception(
            "Failed to read ID for app %s (non UTF-8 encoding). Reverting to executable name.",
            app,
        )
        return app.get_executable()


class XDGMedia(ServiceMedia):
    service = "xdg"
    source = "local"
    size = (64, 64)
    dest_path = os.path.join(settings.CACHE_DIR, "xdg/icons")
    file_patterns = ["%s.png"]


class XDGService(BaseService):
    id = "xdg"
    name = _("Local")
    icon = "linux"
    runner = "linux"
    online = False
    local = True
    medias = {"icon": XDGMedia}

    ignored_games = ("lutris",)
    ignored_executables = ("lutris", "steam")
    ignored_categories = ("Emulator", "Development", "Utility")

    @classmethod
    def iter_xdg_games(cls):
        """Iterates through XDG games only"""
        for app in Gio.AppInfo.get_all():
            if cls._is_importable(app):
                yield app

    @property
    def lutris_games(self):
        """Iterates through Lutris games imported from XDG"""
        for game in get_games_where(runner=XDGGame.runner, installer_slug=XDGGame.installer_slug, installed=1):
            yield game

    @classmethod
    def _is_importable(cls, app):
        """Returns whether a XDG game is importable to Lutris"""
        appid = get_appid(app)
        executable = app.get_executable() or ""
        if any(
            [
                app.get_nodisplay() or app.get_is_hidden(),  # App is hidden
                not executable,  # Check app has an executable
                appid.startswith("net.lutris"),  # Skip lutris created shortcuts
                appid.lower() in map(str.lower, cls.ignored_games),  # game blacklisted
                executable.lower() in cls.ignored_executables,  # exe blacklisted
            ]
        ):
            return False

        # must be in Game category
        categories = app.get_categories() or ""
        categories = list(filter(None, categories.lower().split(";")))
        if "game" not in categories:
            return False

        # contains a blacklisted category
        ignored_categories = set(c.casefold() for c in cls.ignored_categories)
        if any(c for c in categories if c in ignored_categories):
            return False
        return True

    def match_games(self):
        """XDG games aren't on the lutris website"""
        return

    def load(self):
        """Return the list of games stored in the XDG menu."""
        xdg_games = [XDGGame.new_from_xdg_app(app) for app in self.iter_xdg_games()]
        for game in xdg_games:
            game.save()
        return xdg_games

    def generate_installer(self, db_game):
        details = json.loads(db_game["details"])
        return {
            "name": db_game["name"],
            "version": "XDG",
            "slug": db_game["slug"],
            "game_slug": self.get_installed_slug(db_game),
            "runner": self.get_installed_runner_name(db_game),
            "script": {
                "game": {
                    "exe": details["exe"],
                    "args": details["args"],
                    "working_dir": details["path"],
                },
                "system": {"disable_runtime": True},
            },
        }

    def get_installed_runner_name(self, db_game):
        return self.runner

    def get_game_directory(self, installer):
        """Pull install location from installer"""
        return os.path.dirname(installer["script"]["game"]["exe"])


class XDGGame(ServiceGame):
    """XDG game (Linux game with a desktop launcher)"""

    service = "xdg"
    runner = "linux"
    installer_slug = "desktopapp"

    @staticmethod
    def get_app_icon(xdg_app):
        """Return the name of the icon for an XDG app if one if set"""
        icon = xdg_app.get_icon()
        if not icon:
            return ""
        return icon.to_string()

    @classmethod
    def new_from_xdg_app(cls, xdg_app):
        """Create a service game from a XDG entry"""
        service_game = cls()
        service_game.name = xdg_app.get_display_name()
        service_game.icon = cls.get_app_icon(xdg_app)
        service_game.appid = get_appid(xdg_app)
        service_game.slug = cls.get_slug(xdg_app)
        exe, args = cls.get_command_args(xdg_app)
        path = cls.get_desktop_entry_path(xdg_app)
        service_game.details = json.dumps(
            {
                "exe": exe,
                "args": args,
                "path": path,
            }
        )
        return service_game

    @staticmethod
    def get_desktop_entry_path(xdg_app):
        """Retrieve the Path variable from the .desktop file"""

        # I expect we'll only see DesktopAppInfos here, but just in case
        # we get something else, we'll make sure.
        if hasattr(xdg_app, "get_string"):
            return xdg_app.get_string("Path")
        return None

    @staticmethod
    def get_command_args(app):
        """Return a tuple with absolute command path and an argument string"""
        command = shlex.split(app.get_commandline())
        # remove %U etc. and change %% to % in arguments
        args = list(map(lambda arg: re.sub("%[^%]", "", arg).replace("%%", "%"), command[1:]))
        exe = command[0]
        if not exe.startswith("/"):
            exe = system.find_required_executable(exe)
        return exe, subprocess.list2cmdline(args)

    @staticmethod
    def get_slug(xdg_app):
        """Get the slug from the game name"""
        return slugify(xdg_app.get_display_name()) or slugify(get_appid(xdg_app))



# Ubisoft Connect client adapted from the Galaxy integration by Rasmus Luund.
# https://github.com/FriendsOfGalaxy/galaxy-integration-uplay

import json
import time
from datetime import datetime
from gettext import gettext as _

import requests

from lutris.util.log import logger
from lutris.util.ubisoft.consts import CHROME_USERAGENT, CLUB_APPID


def parse_date(date_str):
    date_example = "2022-01-25T05:44:59.192453"
    return datetime.strptime(date_str[: len(date_example)], "%Y-%m-%dT%H:%M:%S.%f")


class UbisoftConnectClient:
    def __init__(self, service):
        self._service = service
        self._auth_lost_callback = None
        self.token = None
        self.session_id = None
        self.refresh_token = None
        self.refresh_time = None
        self.user_id = None
        self.user_name = None
        self.__refresh_in_progress = False
        self._session = requests.session()
        self._session.headers.update(
            {"Authorization": None, "Ubi-AppId": CLUB_APPID, "User-Agent": CHROME_USERAGENT, "Ubi-SessionId": None}
        )

    def close(self):
        # If closing is attempted while plugin is inside refresh workflow then give it a chance to finish it.
        if self.__refresh_in_progress:
            time.sleep(1.5)
        self._session.close()

    def request(self, method, url, *args, **kwargs):
        try:
            return self._session.request(method, url, *args, **kwargs)
        except Exception as ex:
            logger.exception(ex)

    def set_auth_lost_callback(self, callback):
        self._auth_lost_callback = callback

    def is_authenticated(self):
        return self.token is not None

    def _do_request(self, method, *args, **kwargs):
        if not kwargs or "headers" not in kwargs:
            kwargs["headers"] = self._session.headers
        if "add_to_headers" in kwargs:
            for header in kwargs["add_to_headers"]:
                kwargs["headers"][header] = kwargs["add_to_headers"][header]
            kwargs.pop("add_to_headers")

        response = self.request(method, *args, **kwargs)
        result = response.json()
        if "errorCode" in result and "message" in result:
            raise RuntimeError(result["message"])
        return result

    def _do_request_safe(self, method, *args, **kwargs):
        result = {}
        try:
            refresh_needed = False
            if self.refresh_token:
                refresh_needed = self.refresh_time is None or datetime.now() > datetime.fromtimestamp(
                    int(self.refresh_time)
                )
            if refresh_needed:
                self._refresh_auth()
                result = self._do_request(method, *args, **kwargs)
            else:
                try:
                    result = self._do_request(method, *args, **kwargs)
                except Exception:
                    # fallback for another reason than expired time or wrong calculation due to changing time zones
                    logger.debug("Fallback refresh")
                    if not self.refresh_token:
                        logger.warning(
                            "No refresh token present, possibly unchecked remember me when connecting plugin"
                        )
                    self._refresh_auth()
                    result = self._do_request(method, *args, **kwargs)
        except Exception as ex:
            logger.error("Unable to refresh authentication calling auth lost: %s", ex)
            if self._auth_lost_callback:
                self._auth_lost_callback()
            raise RuntimeError(_("Ubisoft authentication has been lost: %s") % ex) from ex
        return result

    def _do_options_request(self):
        self._do_request(
            "options",
            "https://public-ubiservices.ubi.com/v3/profiles/sessions",
            headers={
                "Origin": "https://connect.ubisoft.com",
                "Referer": "https://connect.ubisoft.com/",
                "User-Agent": CHROME_USERAGENT,
            },
        )

    def _refresh_auth(self):
        if self.__refresh_in_progress:
            while self.__refresh_in_progress:
                time.sleep(0.2)
        else:
            self.__refresh_in_progress = True
            try:
                self._refresh_ticket()
                self._service.store_credentials(self.get_credentials())
            except:
                self._refresh_remember_me()
                self._refresh_ticket()
                self._service.store_credentials(self.get_credentials())
            finally:
                self.__refresh_in_progress = False

    def _refresh_remember_me(self):
        self._do_options_request()
        j = self._do_request(
            "post",
            "https://public-ubiservices.ubi.com/v3/profiles/sessions",
            headers={
                "Accept": "*/*",
                "Accept-Encoding": "gzip, deflate, br",
                "Accept-Language": "en-US;en;q=0.5",
                "Authorization": f"rm_v1 t={self.refresh_token}",
                "Content-Type": "application/json",
                "Ubi-AppId": CLUB_APPID,
                "User-Agent": CHROME_USERAGENT,
                "Host": "public-ubiservices.ubi.com",
                "Origin": "https://connect.ubisoft.com",
                "Referer": "https://connect.ubisoft.com",
            },
            json={"rememberMe": True},
        )
        self._handle_authorization_response(j)

    def _refresh_ticket(self):
        self._do_options_request()
        j = self._do_request(
            "put",
            "https://public-ubiservices.ubi.com/v3/profiles/sessions",
            headers={
                "Accept": "*/*",
                "Accept-Encoding": "gzip, deflate, br",
                "Accept-Language": "en-US;en;q=0.5",
                "Authorization": f"Ubi_v1 t={self.token}",
                "Content-Type": "application/json",
                "Ubi-AppId": CLUB_APPID,
                "User-Agent": CHROME_USERAGENT,
                "Host": "public-ubiservices.ubi.com",
                "Origin": "https://connect.ubisoft.com",
                "Referer": "https://connect.ubisoft.com",
            },
        )
        self._handle_authorization_response(j)

    def _handle_authorization_response(self, j):
        if "expiration" in j and "serverTime" in j:
            refresh_time = datetime.now() + (parse_date(j["expiration"]) - parse_date(j["serverTime"]))
            j["refreshTime"] = round(refresh_time.timestamp())
            self.restore_credentials(j)

    def restore_credentials(self, data):
        self.token = data["ticket"]
        self.session_id = data["sessionId"]
        self.user_id = data["userId"]
        if data.get("username"):
            self.user_name = data["username"]
        self.refresh_time = data.get("refreshTime", "0")
        if data.get("rememberMeTicket"):
            self.refresh_token = data["rememberMeTicket"]

        self._session.headers.update({"Authorization": f"Ubi_v1 t={self.token}", "Ubi-SessionId": self.session_id})

    def get_credentials(self):
        creds = {
            "ticket": self.token,
            "sessionId": self.session_id,
            "rememberMeTicket": self.refresh_token,
            "userId": self.user_id,
            "refreshTime": self.refresh_time,
        }

        if self.user_name:
            creds["username"] = self.user_name

        return creds

    def authorise_with_stored_credentials(self, credentials):
        self.restore_credentials(credentials)
        if not self.user_name or not self.user_id:
            user_data = self.get_user_data()
        else:
            user_data = {"username": self.user_name, "userId": self.user_id}
        self.post_sessions()
        self._service.store_credentials(self.get_credentials())
        return user_data

    def authorise_with_local_storage(self, storage_jsons):
        user_data = {}
        tasty_storage_values = ["userId", "nameOnPlatform", "ticket", "rememberMeTicket", "sessionId"]
        for json_ in storage_jsons:
            if json_:  # we do get nulls in this list of json dictionaries sometimes
                for key in json_:
                    if key in tasty_storage_values:
                        user_data[key] = json_[key]

        user_data["userId"] = user_data.pop("userId")
        user_data["username"] = user_data.pop("nameOnPlatform")

        self.restore_credentials(user_data)
        self.post_sessions()
        self._service.store_credentials(self.get_credentials())
        return user_data

    # Deprecated 0.39
    def get_user_data(self):
        return self._do_request_safe("get", f"https://public-ubiservices.ubi.com/v3/users/{self.user_id}")

    def get_friends(self):
        return self._do_request_safe("get", "https://api-ubiservices.ubi.com/v2/profiles/me/friends")

    def get_club_titles(self):
        payload = {
            "operationName": "AllGames",
            "variables": {"owned": True},
            "query": "query AllGames {"
            "viewer {"
            "    id"
            "    ...ownedGamesList"
            "  }"
            "}"
            "fragment gameProps on Game {"
            "  id"
            "  spaceId"
            "  name"
            "  coverUrl"
            "  backgroundUrl"
            "  bannerUrl"
            "}"
            "fragment ownedGameProps on Game {"
            "  ...gameProps"
            "  viewer {"
            "    meta {"
            "      id"
            "      ownedPlatformGroups {"
            "        id"
            "        name"
            "        type"
            "      }"
            "    }"
            "  }"
            "}"
            "fragment ownedGamesList on User {"
            "  ownedGames: games(filterBy: {isOwned: true}) {"
            "    totalCount"
            "    nodes {"
            "      ...ownedGameProps"
            "    }"
            "  }"
            "}",
        }
        payload = json.dumps(payload)
        headers = {"Content-Type": "application/json"}
        return self._do_request_safe(
            "post",
            "https://public-ubiservices.ubi.com/v1/profiles/me/uplay/graphql",
            add_to_headers=headers,
            data=payload,
        )

    def get_game_stats(self, space_id):
        url = f"https://public-ubiservices.ubi.com/v1/profiles/{self.user_id}/statscard?spaceId={space_id}"
        headers = {"Ubi-RequestedPlatformType": "uplay", "Ubi-LocaleCode": "en-GB"}
        try:
            return self._do_request("get", url, add_to_headers=headers)
        except Exception as ex:  # 412: no stats available for this user
            logger.debug(ex)
            return {}

    def get_applications(self, spaces):
        space_string = ",".join(space["spaceId"] for space in spaces)
        return self._do_request_safe("get", f"https://api-ubiservices.ubi.com/v2/applications?spaceIds={space_string}")

    def get_configuration(self):
        response = self._do_request_safe("get", "https://uplaywebcenter.ubi.com/v1/configuration")
        return response.json()

    def post_sessions(self):
        headers = {"Content-Type": "application/json"}
        return self._do_request_safe(
            "post", "https://public-ubiservices.ubi.com/v2/profiles/sessions", add_to_headers=headers
        )

    def get_subscription(self):
        try:
            sub_games = self._do_request("get", "https://api-uplayplusvault.ubi.com/v1/games")
        except Exception:
            logger.info("Uplay plus Subscription not active")
            return None
        return sub_games

    def activate_game(self, activation_id):
        response = self._do_request_safe(
            "post", f"https://api-uplayplusvault.ubi.com/v1/games/activate/{activation_id}"
        )
        return "games" in response



"""Handle Steam configuration"""

import os
from collections import OrderedDict

import requests

from lutris import settings
from lutris.util import system
from lutris.util.log import logger
from lutris.util.steam.steamid import SteamID
from lutris.util.steam.vdfutils import vdf_parse

STEAM_DATA_DIRS = (
    "~/.steam/debian-installation",
    "~/.steam",
    "~/.local/share/steam",
    "~/.local/share/Steam",
    "~/snap/steam/common/.local/share/Steam",
    "~/.steam/steam",
    "~/.var/app/com.valvesoftware.Steam/data/steam",
    "~/.var/app/com.valvesoftware.Steam/data/Steam",
    "/usr/share/steam",
    "/usr/local/share/steam",
)
STEAM_ACCOUNT_SETTING = "active_steam_account"


def get_steam_dir():
    """Main installation directory for Steam"""
    steam_dir = search_in_steam_dirs("steamapps")
    if steam_dir:
        return steam_dir[: -len("steamapps")]


def search_in_steam_dirs(file):
    """Find the (last) file/dir in all the Steam directories"""
    for candidate in STEAM_DATA_DIRS:
        path = system.fix_path_case(os.path.join(os.path.expanduser(candidate), file))
        if path and system.path_exists(path):
            return path


def get_default_acf(appid, name):
    """Return a default configuration usable to
    create a runnable game in Steam"""

    userconfig = OrderedDict()
    userconfig["name"] = name
    userconfig["gameid"] = appid

    appstate = OrderedDict()
    appstate["appID"] = appid
    appstate["Universe"] = "1"
    appstate["StateFlags"] = "1026"
    appstate["installdir"] = name
    appstate["UserConfig"] = userconfig
    return {"AppState": appstate}


def read_user_config():
    config_filename = search_in_steam_dirs("config/loginusers.vdf")
    if not system.path_exists(config_filename):
        return None
    with open(config_filename, "r", encoding="utf-8") as steam_config_file:
        config = vdf_parse(steam_config_file, {})
    return config


def get_config_value(config: dict, key: str):
    """Fetch a value from a configuration in a case insensitive way"""
    keymap = {k.lower(): k for k in config.keys()}
    if key not in keymap:
        logger.warning("Config key %s not found in %s", key, ", ".join(list(config.keys())))
        return
    return config[keymap[key.lower()]]


def get_user_data_dirs():
    """Return the list of available Steam user config directories (using a SteamID32)
    and the base path where the settings are located (Returns the 1st location found)
    """
    for steam_dir in STEAM_DATA_DIRS:
        userdata_path = os.path.join(os.path.expanduser(steam_dir), "userdata")
        if not os.path.exists(userdata_path):
            continue
        user_ids = [f for f in os.listdir(userdata_path) if f.isnumeric()]
        if user_ids:
            return userdata_path, user_ids
    return "", []


def get_steam_users() -> list:
    """Return a list of available Steam users.
    Most recently used account is 1st in the list."""
    steam_users = []
    user_config = read_user_config()
    if not user_config or "users" not in user_config:
        return []
    most_recent = None
    for steam_id, account in user_config["users"].items():
        account["steamid64"] = steam_id
        if get_config_value(account, "mostrecent") == "1":
            most_recent = account
        else:
            steam_users.append(account)
    if most_recent:
        steam_users = [most_recent] + steam_users
    return steam_users


def get_active_steamid64() -> str:
    """Return the currently active Steam ID"""
    steam_ids = [u["steamid64"] for u in get_steam_users()]
    active_steam_id = settings.read_setting(STEAM_ACCOUNT_SETTING)
    if active_steam_id in steam_ids:
        return active_steam_id
    if steam_ids:
        return steam_ids[0]
    return ""


def convert_steamid64_to_steamid32(steamid64: str) -> str:
    """Return the 32-bit variant of SteamIDs, used for folder names in userdata"""
    if not steamid64.isnumeric():
        return ""
    steam_id = SteamID.from_steamid64(int(steamid64))
    return str(steam_id.get_32_bit_community_id())


def get_steam_library(steamid):
    """Return the list of games owned by a SteamID"""
    if not steamid:
        raise ValueError("Missing SteamID")
    steam_games_url = (
        "https://api.steampowered.com/"
        "IPlayerService/GetOwnedGames/v0001/"
        "?key={}&steamid={}&format=json&include_appinfo=1"
        "&include_played_free_games=1".format(settings.STEAM_API_KEY, steamid)
    )
    response = requests.get(steam_games_url, timeout=30)
    if response.status_code > 400:
        logger.error("Invalid response from steam: %s", response)
        return []
    json_data = response.json()
    response = json_data["response"]
    if not response:
        logger.info("No games in response of %s", steam_games_url)
        return []
    if "games" in response:
        return response["games"]
    if "game_count" in response and response["game_count"] == 0:
        return []
    logger.error("Weird response: %s", json_data)
    return []


def read_config(steam_data_dir):
    """Read the Steam configuration and return it as an object"""

    def get_entry_case_insensitive(config_dict, path):
        for key, _value in config_dict.items():
            if key.lower() == path[0].lower():
                if len(path) <= 1:
                    return config_dict[key]

                return get_entry_case_insensitive(config_dict[key], path[1:])
        raise KeyError(path[0])

    if not steam_data_dir:
        return None
    config_filename = os.path.join(steam_data_dir, "config/config.vdf")
    if not system.path_exists(config_filename):
        return None
    with open(config_filename, "r", encoding="utf-8") as steam_config_file:
        config = vdf_parse(steam_config_file, {})
    try:
        return get_entry_case_insensitive(config, ["InstallConfigStore", "Software", "Valve", "Steam"])
    except KeyError as ex:
        logger.error("Steam config %s is empty: %s", config_filename, ex)


def read_library_folders(steam_data_dir):
    """Read the Steam Library Folders config and return it as an object"""

    def get_entry_case_insensitive(library_dict, path):
        for key, value in library_dict.items():
            if key.lower() == path[0].lower():
                if len(path) <= 1:
                    return value
                return get_entry_case_insensitive(library_dict[key], path[1:])
            raise KeyError(path[0])

    if not steam_data_dir:
        return None
    library_filename = os.path.join(steam_data_dir, "config/libraryfolders.vdf")
    if not system.path_exists(library_filename):
        return None
    with open(library_filename, "r", encoding="utf-8") as steam_library_file:
        library = vdf_parse(steam_library_file, {})
        # The contentstatsid key is unused and causes problems when looking for library paths.
        library["libraryfolders"].pop("contentstatsid", None)
    try:
        return get_entry_case_insensitive(library, ["libraryfolders"])
    except KeyError as ex:
        logger.error("Steam libraryfolders %s is empty: %s", library_filename, ex)


def get_steam_config():
    """Return the "Steam" part of Steam's config.vdf as a dict."""
    return read_config(get_steam_dir())


def get_library_config():
    """Return the "libraryfolders" part of Steam's libraryfolders.vdf as a dict"""
    return read_library_folders(get_steam_dir())


def get_steamapps_dirs():
    """Return a list of the Steam library main + custom folders."""
    dirs = []
    # Extra colon-separated compatibility tools dirs environment variable
    if "STEAM_EXTRA_COMPAT_TOOLS_PATHS" in os.environ:
        dirs += os.getenv("STEAM_EXTRA_COMPAT_TOOLS_PATHS").split(":")
    # Main steamapps dir and compatibilitytools.d dir
    for data_dir in STEAM_DATA_DIRS:
        for _dir in ["steamapps", "compatibilitytools.d"]:
            abs_dir = os.path.join(os.path.expanduser(data_dir), _dir)
            abs_dir = system.fix_path_case(abs_dir)
            if abs_dir and os.path.isdir(abs_dir):
                dirs.append(abs_dir)

    # Custom dirs
    steam_config = get_steam_config()
    if steam_config:
        i = 1
        while "BaseInstallFolder_%s" % i in steam_config:
            path = steam_config["BaseInstallFolder_%s" % i] + "/steamapps"
            path = system.fix_path_case(path)
            if path and os.path.isdir(path):
                dirs.append(path)
            i += 1

    # New Custom dirs
    library_config = get_library_config()
    if library_config:
        paths = []
        for entry in library_config.values():
            if "mounted" in entry:
                if entry.get("path") and entry.get("mounted") == "1":
                    path = system.fix_path_case(entry.get("path") + "/steamapps")
                    paths.append(path)
            else:
                path = system.fix_path_case(entry.get("path") + "/steamapps")
                paths.append(path)
        for path in paths:
            if path and os.path.isdir(path):
                dirs.append(path)
    return system.list_unique_folders(dirs)



from gettext import gettext as _

from lutris.config import LutrisConfig
from lutris.gui.config.game_common import GameDialogCommon


class AddGameDialog(GameDialogCommon):
    """Add game dialog class."""

    def __init__(self, parent, game=None, runner=None):
        super().__init__(_("Add a new game"), config_level="game", parent=parent)
        self.game = game
        self.saved = False

        self.lutris_config = LutrisConfig(
            runner_slug=game.runner_name if game else runner,
            level="game",
        )
        self.build_notebook()
        self.build_tabs()
        self.info_box.grab_focus()
        self.show_all()



# pylint: disable=no-member
from gettext import gettext as _
from typing import Dict, Set

from gi.repository import Gtk

from lutris.database import categories as categories_db
from lutris.database import games as games_db
from lutris.game import GAME_UPDATED, Game
from lutris.gui.dialogs import QuestionDialog, SavableModelessDialog
from lutris.util.strings import get_natural_sort_key


class EditCategoryGamesDialog(SavableModelessDialog):
    """Games assigned to category dialog."""

    def __init__(self, parent, category):
        super().__init__(_("Configure %s") % category["name"], parent=parent, border_width=10)

        self.category = category["name"]
        self.category_id = category["id"]
        self.available_games = sorted(
            [Game(x["id"]) for x in games_db.get_games()], key=lambda g: (g.is_installed, get_natural_sort_key(g.name))
        )
        self.category_games = {
            game_id: Game(game_id) for game_id in categories_db.get_game_ids_for_categories([self.category])
        }
        self.grid = Gtk.Grid()

        self.set_default_size(500, 350)

        self.vbox.set_homogeneous(False)
        self.vbox.set_spacing(10)
        name_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        name_label = Gtk.Label(_("Name"))
        self.name_entry = Gtk.Entry()
        self.name_entry.set_text(self.category)
        name_box.pack_start(name_label, False, False, 0)
        name_box.pack_start(self.name_entry, True, True, 0)

        self.vbox.pack_start(name_box, False, False, 0)
        self.vbox.pack_start(self._create_games_checkboxes(), True, True, 0)

        delete_button = self.add_styled_button(Gtk.STOCK_DELETE, Gtk.ResponseType.NONE, css_class="destructive-action")
        delete_button.connect("clicked", self.on_delete_clicked)

        self.show_all()

    def _create_games_checkboxes(self):
        frame = Gtk.Frame()
        sw = Gtk.ScrolledWindow()
        row = Gtk.VBox()
        category_games_names = sorted([x.name for x in self.category_games.values()])
        for game in self.available_games:
            label = game.name
            checkbutton_option = Gtk.CheckButton(label)
            if label in category_games_names:
                checkbutton_option.set_active(True)
            self.grid.attach_next_to(checkbutton_option, None, Gtk.PositionType.BOTTOM, 3, 1)

        row.pack_start(self.grid, True, True, 0)
        sw.add_with_viewport(row)
        frame.add(sw)
        return frame

    def on_delete_clicked(self, _button):
        dlg = QuestionDialog(
            {
                "title": _("Do you want to delete the category '%s'?") % self.category,
                "question": _(
                    "This will permanently destroy the category, but the games themselves will not be deleted."
                ),
                "parent": self,
            }
        )
        if dlg.result == Gtk.ResponseType.YES:
            for game in self.category_games.values():
                game.remove_category(self.category)
            categories_db.remove_category(self.category_id)
            self.destroy()

    def _get_game(self, game_id: str) -> Game:
        if game_id in self.category_games:
            return self.category_games[game_id]
        else:
            return Game(game_id)

    def on_save(self, _button: Gtk.Button) -> None:
        """Save game info and destroy widget."""
        old_name: str = self.category
        new_name: str = categories_db.strip_category_name(self.name_entry.get_text())
        category_games_ids: Set[str] = set(self.category_games.keys())

        # Work out which games hae been added or removed from the category
        unchecked_game_ids: Set[str] = set()
        checked_game_ids: Set[str] = set()
        updated_games: Dict[str, Game] = {}

        for game_checkbox in self.grid.get_children():
            label = game_checkbox.get_label()
            game_id = games_db.get_game_by_field(label, "name")["id"]
            if game_checkbox.get_active():
                checked_game_ids.add(game_id)
            else:
                unchecked_game_ids.add(game_id)

        added_game_ids = checked_game_ids - category_games_ids
        removed_game_ids = unchecked_game_ids & category_games_ids

        # Rename the category if required, and if this is not a merge
        if new_name and old_name != new_name:
            if categories_db.is_reserved_category(new_name):
                raise RuntimeError(_("'%s' is a reserved category name.") % new_name)

            if new_name in (c["name"] for c in categories_db.get_categories()):
                dlg = QuestionDialog(
                    {
                        "title": _("Merge the category '%s' into '%s'?") % (old_name, new_name),
                        "question": _(
                            "If you rename this category, it will be combined with '%s'. Do you want to merge them?"
                        )
                        % new_name,
                        "parent": self,
                    }
                )
                if dlg.result != Gtk.ResponseType.YES:
                    return

                # To merge, remove every categry and add them all to the
                # other one.
                removed_game_ids = category_games_ids
                added_game_ids = checked_game_ids
            else:
                categories_db.redefine_category(self.category_id, new_name)
                old_name = new_name

                updated_games = {
                    game_id: game
                    for game_id, game in self.category_games.items()
                    if game_id not in added_game_ids and game_id not in removed_game_ids
                }

        # Apply category changes and fire GAME_UPDATED as needed after everything
        for game_id in added_game_ids:
            game = self._get_game(game_id)
            game.add_category(new_name, no_signal=True)
            updated_games[game_id] = game

        for game_id in removed_game_ids:
            game = self._get_game(game_id)
            game.remove_category(old_name, no_signal=True)
            updated_games[game_id] = game

        for game in updated_games.values():
            GAME_UPDATED.fire(game)

        self.destroy()



# pylint: disable=no-member
import os
from typing import Any, Callable, Dict, Iterable, List, Optional, Set

from gi.repository import GObject, Gtk  # type: ignore

from lutris.gui.widgets.gi_composites import GtkTemplate
from lutris.gui.widgets.progress_box import ProgressBox
from lutris.util import datapath
from lutris.util.jobs import AsyncCall
from lutris.util.log import logger


@GtkTemplate(ui=os.path.join(datapath.get(), "ui", "download-queue.ui"))
class DownloadQueue(Gtk.ScrolledWindow):
    """This class is a widget that displays a stack of progress boxes, which you can create
    and destroy with its methods."""

    __gsignals__ = {
        "download-completed": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    __gtype_name__ = "DownloadQueue"

    download_box: Gtk.Box = GtkTemplate.Child()  # type: ignore

    CompletionFunction = Callable[[Any], None]
    ErrorFunction = Callable[[Exception], None]

    def __init__(self, revealer: Gtk.Revealer, **kwargs):
        super().__init__(**kwargs)
        self.revealer = revealer
        self.init_template()

        self.running_operation_names: Set[str] = set()
        self.progress_boxes: Dict[ProgressBox.ProgressFunction, ProgressBox] = {}

        try:
            # GTK 3.22 is required for this, but if this fails we can still run.
            # The download area comes out too small, but it's usable.
            self.set_max_content_height(250)
            self.set_propagate_natural_height(True)
        except AttributeError:
            pass

    @property
    def is_empty(self):
        """True if the queue has no progress boxes in it."""
        return not bool(self.progress_boxes)

    def add_progress_box(self, progress_function: ProgressBox.ProgressFunction) -> ProgressBox:
        """Adds a progress box to the queue; it will display the progress indicated by
        the progress_function, which is called immediately to initialize the box and
        then polled to update it. Returns the new progress box.

        The progres-box is removed when its function returns ProgressInfo.ended(), or when
        you call remove_progress_box().

        If called with a progress_function that has a box already, this method returns
        that box instead of creating one."""

        def check_progress():
            progress_info = progress_function()

            if progress_info.has_ended:
                self.remove_progress_box(progress_function)
                return progress_info

            if progress_info.label_markup:
                progress_info.label_markup = "<span size='10000'>%s</span>" % progress_info.label_markup

            return progress_info

        progress_box = self.progress_boxes.get(progress_function)
        if progress_box:
            progress_box.update_progress()
            return progress_box

        progress_box = ProgressBox(check_progress, visible=False, margin=6)
        progress_box.update_progress()

        self.progress_boxes[progress_function] = progress_box
        self.download_box.pack_start(progress_box, False, False, 0)
        progress_box.show()
        self.revealer.set_reveal_child(True)
        return progress_box

    def remove_progress_box(self, progress_function: ProgressBox.ProgressFunction) -> None:
        """Removes and destroys the progress box created for the progress_function given,
        if any is present."""
        progress_box = self.progress_boxes.get(progress_function)
        if progress_box:
            del self.progress_boxes[progress_function]
            progress_box.destroy()
            if not self.progress_boxes:
                self.revealer.set_reveal_child(False)

    def start(
        self,
        operation: Callable[[], Any],
        progress_function: ProgressBox.ProgressFunction,
        completion_function: Optional[CompletionFunction] = None,
        error_function: Optional[ErrorFunction] = None,
        operation_name: Optional[str] = None,
    ) -> bool:
        """Runs 'operation' on a thread, while displaying a progress bar. The 'progress_function'
        controls this progress bar, and it is removed when the 'operation' completes.

        If 'operation_name' is given, it is added to self.running_operation_names while
        the 'operation' runs. If the name is present already, this method does nothing
        but returns False. If the worker thread has started, this returns True.

        Args:
            operation:              Called on a worker thread
            progress_function:      Called on the main thread for progress status
            completion_function:    Called on the main thread on completion, with result
            error_function:         Called on the main threa don error, with exception
            operation_name:         Name of operation, to prevent duplicate queued work."""

        return self.start_multiple(
            operation,
            [progress_function],
            completion_function=completion_function,
            error_function=error_function,
            operation_names=[operation_name] if operation_name else None,
        )

    def start_multiple(
        self,
        operation: Callable[[], Any],
        progress_functions: Iterable[ProgressBox.ProgressFunction],
        completion_function: Optional[CompletionFunction] = None,
        error_function: Optional[ErrorFunction] = None,
        operation_names: Optional[List[str]] = None,
    ) -> bool:
        """Runs 'operation' on a thread, while displaying a set of progress bars. The
        'progress_functions' control these progress bars, and they are removed when the
        'operation' completes.

        If 'operation_names' is given, they are added to self.running_operation_names while
        the 'operation' runs. If any name is present already, this method does nothing
        but returns False. If the worker thread has started, this returns True.

        Args:
            operation:              Called on a worker thread
            progress_functions:     Called on the main thread for progress status
            completion_function:    Called on the main thread on completion, with result
            error_function:         Called on the main threa don error, with exception
            operation_names:        Names of operations, to prevent duplicate queued work."""

        if operation_names:
            if not self.running_operation_names.isdisjoint(operation_names):
                return False
            self.running_operation_names.update(operation_names)
        else:
            operation_names = []

        # Must capture the functions, since in earlier (<3.8) Pythons functions do not provide
        # value equality, so we need to make sure we're always using what we started with.
        captured_functions = list(progress_functions)

        for f in captured_functions:
            self.add_progress_box(f)

        def completion_callback(result, error):
            for to_end in captured_functions:
                self.remove_progress_box(to_end)

            self.running_operation_names.difference_update(operation_names)

            if error:
                logger.exception("Failed to execute download-queue function: %s", error)
                if error_function:
                    error_function(error)
            elif completion_function:
                completion_function(result)
            self.emit("download-completed")

        AsyncCall(operation, completion_callback)
        return True



"""DXVK helper module"""

import os

from lutris.util.graphics import vkquery
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.wine.dll_manager import DLLManager

REQUIRED_VULKAN_API_VERSION = vkquery.vk_make_version(1, 3, 0)


class DXVKManager(DLLManager):
    name = "dxvk"
    human_name = "DXVK"
    managed_dlls = ("dxgi", "d3d11", "d3d10core", "d3d9", "d3d8")
    releases_url = "https://api.github.com/repos/lutris/dxvk/releases"

    def can_enable(self):
        if os.environ.get("LUTRIS_NO_VKQUERY"):
            return True
        return LINUX_SYSTEM.is_vulkan_supported()

    def is_recommended_version(self, version):
        # DXVK 2.x and later require Vulkan 1.3, so if that is lacking
        # we default to 1.x.
        if os.environ.get("LUTRIS_NO_VKQUERY"):
            return True
        vulkan_api_version = vkquery.get_expected_api_version()
        if vulkan_api_version and vulkan_api_version < REQUIRED_VULKAN_API_VERSION:
            return version.startswith("v1.")
        return super().is_recommended_version(version)

    @staticmethod
    def is_managed_dll(dll_path):
        """Check if a given DLL path is provided by the component

        Very basic check to see if a dll contains the string "dxvk".
        """
        try:
            with open(dll_path, "rb") as file:
                prev_block_end = b""
                while True:
                    block = file.read(2 * 1024 * 1024)  # 2 MiB
                    if not block:
                        break
                    if b"dxvk" in prev_block_end + block[:4]:
                        return True
                    if b"dxvk" in block:
                        return True

                    prev_block_end = block[-4:]
        except OSError:
            pass
        return False



from gettext import gettext as _

from gi.repository import Gtk

from lutris.gui.dialogs import ModalDialog
from lutris.gui.widgets.download_progress_box import DownloadProgressBox
from lutris.util.downloader import Downloader


class DownloadDialog(ModalDialog):
    """Dialog showing a download in progress."""

    def __init__(self, url=None, dest=None, title=None, label=None, downloader=None, parent=None):
        super().__init__(title=title or _("Downloading file"), parent=parent, border_width=10)
        self.set_size_request(485, 104)
        self.dialog_progress_box = DownloadProgressBox(
            url=url, dest=dest, title=label or _("Downloading %s") % url, downloader=downloader
        )

        self.dialog_progress_box.connect("complete", self.download_complete)
        self.dialog_progress_box.connect("cancel", self.download_cancelled)

        self.get_content_area().add(self.dialog_progress_box)
        self.show_all()
        self.dialog_progress_box.start()

    @property
    def downloader(self) -> Downloader:
        return self.dialog_progress_box.downloader

    def download_complete(self, _widget, _data):
        self.response(Gtk.ResponseType.OK)

    def download_cancelled(self, _widget):
        self.response(Gtk.ResponseType.CANCEL)

    def on_response(self, dialog, response):
        if response in (Gtk.ResponseType.DELETE_EVENT, Gtk.ResponseType.CANCEL):
            self.dialog_progress_box.cancel_download()
        super().on_response(dialog, response)



import logging as log
import math

import yaml

from lutris.util.ubisoft.consts import UBISOFT_CONFIGURATIONS_BLACKLISTED_NAMES


class UbisoftParser(object):
    def __init__(self):
        self.configuration_raw = None
        self.ownership_raw = None
        self.settings_raw = None
        self.parsed_games = None

    def _convert_data(self, data):
        # calculate object size (konrad's formula)
        if data > 256 * 256:
            data = data - (128 * 256 * math.ceil(data / (256 * 256)))
            data = data - (128 * math.ceil(data / 256))
        else:
            if data > 256:
                data = data - (128 * math.ceil(data / 256))
        return data

    def _parse_configuration_header(self, header, second_eight=False):
        try:
            offset = 1
            multiplier = 1
            record_size = 0
            tmp_size = 0

            if second_eight:
                while header[offset] != 0x08 or (header[offset] == 0x08 and header[offset + 1] == 0x08):
                    record_size += header[offset] * multiplier
                    multiplier *= 256
                    offset += 1
                    tmp_size += 1
            else:
                while header[offset] != 0x08 or record_size == 0:
                    record_size += header[offset] * multiplier
                    multiplier *= 256
                    offset += 1
                    tmp_size += 1

            record_size = self._convert_data(record_size)

            offset += 1  # skip 0x08

            # look for launch_id
            multiplier = 1
            launch_id = 0

            while header[offset] != 0x10 or header[offset + 1] == 0x10:
                launch_id += header[offset] * multiplier
                multiplier *= 256
                offset += 1

            launch_id = self._convert_data(launch_id)

            offset += 1  # skip 0x10

            multiplier = 1
            launch_id_2 = 0
            while header[offset] != 0x1A or (header[offset] == 0x1A and header[offset + 1] == 0x1A):
                launch_id_2 += header[offset] * multiplier
                multiplier *= 256
                offset += 1

            launch_id_2 = self._convert_data(launch_id_2)

            # if object size is smaller than 128b, there might be a chance that secondary size will not occupy 2b
            if record_size - offset < 128 <= record_size:
                tmp_size -= 1
                record_size += 1

            # we end up in the middle of header, return values normalized
            # to end of record as well real yaml size and game launch_id
            return record_size - offset, launch_id, launch_id_2, offset + tmp_size + 1
        except:
            # something went horribly wrong, do not crash it,
            # just return 0s, this way it will be handled later in the code
            # 10 is to step a little in configuration file in order to find next game
            return 0, 0, 0, 10

    def _parse_ownership_header(self, header):
        offset = 1
        multiplier = 1
        record_size = 0
        tmp_size = 0
        if header[offset - 1] == 0x0A:
            while header[offset] != 0x08 or record_size == 0:
                record_size += header[offset] * multiplier
                multiplier *= 256
                offset += 1
                tmp_size += 1

            record_size = self._convert_data(record_size)

            offset += 1  # skip 0x08

            # look for launch_id
            multiplier = 1
            launch_id = 0

            while header[offset] != 0x10 or header[offset + 1] == 0x10:
                launch_id += header[offset] * multiplier
                multiplier *= 256
                offset += 1

            launch_id = self._convert_data(launch_id)

            offset += 1  # skip 0x10

            multiplier = 1
            launch_id_2 = 0
            while header[offset] != 0x22:
                launch_id_2 += header[offset] * multiplier
                multiplier *= 256
                offset += 1

            launch_id_2 = self._convert_data(launch_id_2)
            return launch_id, launch_id_2, record_size + tmp_size + 1
        else:
            return None, None, None

    def _parse_configuration(self):
        configuration_content = self.configuration_raw
        global_offset = 0
        records = {}
        if configuration_content:
            try:
                while global_offset < len(configuration_content):
                    data = configuration_content[global_offset:]
                    object_size, install_id, launch_id, header_size = self._parse_configuration_header(data)

                    launch_id = install_id if launch_id == 0 or launch_id == install_id else launch_id

                    if object_size > 500:
                        records[install_id] = {
                            "size": object_size,
                            "offset": global_offset + header_size,
                            "install_id": install_id,
                            "launch_id": launch_id,
                        }
                    global_offset_tmp = global_offset
                    global_offset += object_size + header_size

                    if global_offset < len(configuration_content) and configuration_content[global_offset] != 0x0A:
                        object_size, _, _, header_size = self._parse_configuration_header(data, True)
                        global_offset = global_offset_tmp + object_size + header_size
            except:
                log.exception("parse_configuration failed with exception. Possibly 'configuration' file corrupted")
        return records

    def _parse_ownership(self):
        ownership_content = self.ownership_raw
        global_offset = 0x108
        records = []
        try:
            while global_offset < len(ownership_content):
                data = ownership_content[global_offset:]
                launch_id, launch_id2, record_size = self._parse_ownership_header(data)
                if launch_id:
                    records.append(launch_id)
                    if launch_id2 != launch_id:
                        records.append(launch_id2)
                    global_offset += record_size
                else:
                    break
        except:
            log.exception("parse_ownership failed with exception. Possibly 'ownership' file corrupted")
            return []
        return records

    def _parse_user_settings(self):
        def get_game_id(data, rec_size):
            i = 0
            multiplier = 1
            game_id = 0
            while i < rec_size:
                game_id += data[i] * multiplier
                multiplier *= 256
                i += 1

            if game_id > 256 * 256:
                game_id -= 128 * 256 * math.ceil(game_id / (256 * 256))
                game_id -= 128 * math.ceil(game_id / 256)
            else:
                if game_id > 256:
                    game_id -= 128 * math.ceil(game_id / 256)

            return str(game_id)

        """
            0Ah - file start
            0Ah - hidden games records
                00h -> no hidden games
                !00h -> hidden games (hidden games total entry size)
                    0Ah - SEPARATOR
                    03h - RECORD SIZE
                    08h - SEPARATOR
                    [RECORD_SIZE-1] -> game_ID in konrad's format
                    [..]
            12h -> fav games records
                00h -> no fav games
                !00h -> fav games (fav games total entry size)
                    0Ah - SEPARATOR
                    03h - RECORD SIZE
                    08h - SEPARATOR
                    [RECORD_SIZE-1] -> game_ID in konrad's format
                    [..]
        """
        global_offset = 1
        fav = set()
        hidden = set()
        data = self.settings_raw
        if data[global_offset] != 0:
            buffer = int(data[global_offset])
            fav_records = data[global_offset + 1 : global_offset + buffer + 1]
        else:
            fav_records = []

        global_offset = len(fav_records) + 3
        if data[global_offset] != 0:
            buffer = int(data[global_offset])
            hidden_records = data[global_offset + 1 : global_offset + buffer + 1]
        else:
            hidden_records = []

        pos = 0
        while pos < len(fav_records):
            rec_size = fav_records[pos + 1] - 1
            rec_data = fav_records[pos + 3 : pos + 3 + rec_size]
            fav.add(get_game_id(rec_data, rec_size))
            pos += rec_size + 3

        pos = 0
        while pos < len(hidden_records):
            rec_size = hidden_records[pos + 1] - 1
            rec_data = hidden_records[pos + 3 : pos + 3 + rec_size]
            hidden.add(get_game_id(rec_data, rec_size))
            pos += rec_size + 3

        return fav, hidden

    def _get_field_from_yaml(self, game_yaml, field="name"):
        field_value = ""

        if field in game_yaml["root"]:
            field_value = game_yaml["root"][field]
        # Fallback 1
        if field == "name" and field_value.lower() in UBISOFT_CONFIGURATIONS_BLACKLISTED_NAMES:
            if "installer" in game_yaml["root"] and "game_identifier" in game_yaml["root"]["installer"]:
                field_value = game_yaml["root"]["installer"]["game_identifier"]
        # Fallback 2
        if field_value.lower() in UBISOFT_CONFIGURATIONS_BLACKLISTED_NAMES:
            if (
                "localizations" in game_yaml
                and "default" in game_yaml["localizations"]
                and field_value in game_yaml["localizations"]["default"]
            ):
                field_value = game_yaml["localizations"]["default"][field_value]
        return field_value

    def _get_steam_game_properties_from_yaml(self, game_yaml):
        path = ""
        third_party_id = ""
        if "third_party_steam" in game_yaml["root"]["start_game"]:
            path = game_yaml["root"]["start_game"]["third_party_steam"]["game_installation_status_register"]
            third_party_id = game_yaml["root"]["start_game"]["third_party_steam"]["steam_app_id"]
        elif "steam" in game_yaml["root"]["start_game"]:
            path = game_yaml["root"]["start_game"]["steam"]["game_installation_status_register"]
            third_party_id = game_yaml["root"]["start_game"]["steam"]["steam_app_id"]
        return path, third_party_id

    def _get_registry_properties_from_yaml(self, game_yaml):
        try:
            registry_path = game_yaml["root"]["start_game"]["online"]["executables"][0]["working_directory"]["register"]
            try:
                exe = game_yaml["root"]["start_game"]["online"]["executables"][0]["path"]["relative"]
            except KeyError:
                exe = game_yaml["root"]["start_game"]["online"]["executables"][0]["path"]["register"]
            return registry_path, exe
        except (KeyError, IndexError):
            return None, None

    def _parse_game(self, game_yaml, install_id, launch_id):
        space_id = ""
        launch_id = str(launch_id)
        install_id = str(install_id)
        if "space_id" in game_yaml["root"]:
            space_id = game_yaml["root"]["space_id"]
        game_name = self._get_field_from_yaml(game_yaml, "name")
        thumb_image = self._get_field_from_yaml(game_yaml, "thumb_image")
        registry_path, exe = self._get_registry_properties_from_yaml(game_yaml)
        return {
            "spaceId": space_id,
            "launchId": launch_id,
            "installId": install_id,
            "name": game_name,
            "thumbImage": thumb_image,
            "registryPath": registry_path,
            "exe": exe,
        }

    def _iter_configuration_items(self):
        configuration_records = self._parse_configuration()
        for _, game in configuration_records.items():
            if game["size"]:
                yield game

    def _parse_configuration_item(self, conf_item):
        stream = self.configuration_raw[conf_item["offset"] : conf_item["offset"] + conf_item["size"]].decode(
            "utf8", errors="ignore"
        )
        if stream and "start_game" in stream:
            return yaml.load(stream.replace("\t", " "), Loader=yaml.FullLoader)

    def parse_games(self, configuration_data):
        self.configuration_raw = configuration_data
        for game in self._iter_configuration_items():
            yaml_object = self._parse_configuration_item(game)
            if yaml_object:
                yield self._parse_game(yaml_object, game["install_id"], game["launch_id"])

    def get_owned_local_games(self, ownership_data):
        self.ownership_raw = ownership_data
        return self._parse_ownership()

    def get_game_tags(self, settings_data):
        self.settings_raw = settings_data
        return self._parse_user_settings()



# Standard Library
import os
from gettext import gettext as _

from lutris.exceptions import MissingBiosError, MissingGameExecutableError

# Lutris Modules
from lutris.runners.runner import Runner
from lutris.util import system


class jzintv(Runner):
    human_name = _("jzIntv")
    description = _("Intellivision Emulator")
    platforms = [_("Intellivision")]
    runner_executable = "jzintv/bin/jzintv"
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("ROM file"),
            "default_path": "game_path",
            "help": _(
                "The game data, commonly called a ROM image. \n"
                "Supported formats: ROM, BIN+CFG, INT, ITV \n"
                "The file extension must be lower-case."
            ),
        }
    ]
    runner_options = [
        {
            "option": "bios_path",
            "type": "directory",
            "label": _("Bios location"),
            "help": _(
                "Choose the folder containing the Intellivision BIOS "
                "files (exec.bin and grom.bin).\n"
                "These files contain code from the original hardware "
                "necessary to the emulation."
            ),
        },
        {"option": "fullscreen", "type": "bool", "section": _("Graphics"), "label": _("Fullscreen")},
        {
            "option": "resolution",
            "type": "choice",
            "section": _("Graphics"),
            "label": _("Resolution"),
            "choices": (
                ("320 x 200", "0"),
                ("640 x 480", "1"),
                ("800 x 400", "5"),
                ("800 x 600", "2"),
                ("1024 x 768", "3"),
                ("1680 x 1050", "4"),
                ("1600 x 1200", "6"),
            ),
            "default": "0",
        },
    ]

    def play(self):
        """Run Intellivision game"""
        arguments = self.get_command()

        selected_resolution = self.runner_config.get("resolution")
        if selected_resolution:
            arguments = arguments + ["-z%s" % selected_resolution]

        if self.runner_config.get("fullscreen"):
            arguments = arguments + ["-f"]

        bios_path = self.runner_config.get("bios_path", "")
        if system.path_exists(bios_path):
            arguments.append("--execimg=%s/exec.bin" % bios_path)
            arguments.append("--gromimg=%s/grom.bin" % bios_path)
        else:
            raise MissingBiosError()
        rom_path = self.game_config.get("main_file") or ""
        if not system.path_exists(rom_path):
            raise MissingGameExecutableError(filename=rom_path)
        romdir = os.path.dirname(rom_path)
        romfile = os.path.basename(rom_path)
        arguments += ["--rom-path=%s/" % romdir]
        arguments += [romfile]
        return {"command": arguments}



"""Steam game library watcher"""

# Third Party Libraries
# pylint: disable=too-few-public-methods
from gi.repository import Gio, GLib

# Lutris Modules
from lutris.util.log import logger


class SteamWatcher:
    """Watches a Steam library folder and notify changes"""

    def __init__(self, steamapps_paths, callback=None):
        self.monitors = []
        self.callback = callback
        for steam_path in steamapps_paths:
            path = Gio.File.new_for_path(steam_path)
            try:
                monitor = path.monitor_directory(Gio.FileMonitorFlags.NONE)
                logger.debug("Watching Steam folder %s", steam_path)
                monitor.connect("changed", self._on_directory_changed)
                self.monitors.append(monitor)
            except GLib.Error as ex:
                logger.exception("Failed to monitor Steam folder %s: %s", steam_path, ex)

    def _on_directory_changed(self, _monitor, _file, _other_file, event_type):
        path = _file.get_path()
        if path is None or not path.endswith(".acf"):
            return
        if self.callback:
            self.callback(event_type, path)



"""Widget generators and their signal handlers"""

import os
from abc import abstractmethod

# Standard Library
# pylint: disable=no-member,too-many-public-methods
from gettext import gettext as _
from typing import Any, Dict, Optional

# Third Party Libraries
from gi.repository import Gtk, Pango

# Lutris Modules
from lutris import settings, sysoptions
from lutris.config import LutrisConfig
from lutris.game import Game
from lutris.gui.config.widget_generator import WidgetGenerator
from lutris.gui.widgets.common import VBox
from lutris.runners import InvalidRunnerError, import_runner
from lutris.util.log import logger
from lutris.util.wine.wine import clear_wine_version_cache


def set_option_wrapper_style_class(wrapper: Gtk.Widget, class_name: Optional[str]):
    """Sets a particular CSS class on a wrapper, and removes any other classes that start
    with 'option-wrapper-' so there's only one o these classes."""
    style_context = wrapper.get_style_context()

    for cls in style_context.list_classes():
        if cls.startswith("option-wrapper-") and cls != class_name:
            style_context.remove_class(cls)

    if class_name:
        style_context.add_class(class_name)


class AdvancedSettingsBox(VBox):
    """Intermediate vbox class for expsoing the Advanced Visibility options"""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._advanced_visibility = False

    @property
    def advanced_visibility(self):
        return self._advanced_visibility

    @advanced_visibility.setter
    def advanced_visibility(self, value):
        """Sets the visibility of every 'advanced' option and every section that
        contains only 'advanced' options."""
        self._advanced_visibility = value
        self.update_widgets()

    @abstractmethod
    def update_widgets(self):
        """Updates widgets on visibility change; this method must be
        implemented by a subclass."""
        raise NotImplementedError()


class ConfigBox(AdvancedSettingsBox):
    """Dynamically generate a vbox built upon on a python dict."""

    config_section = NotImplemented

    def __init__(self, config_level: str, lutris_config: LutrisConfig, game: Game = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.options = []
        self.config_level = config_level
        self.lutris_config = lutris_config
        self.game = game
        self.config = None
        self.raw_config = None
        self.files = []
        self.files_list_store = None
        self._widget_generator = None
        self._filter = ""
        self._filter_text = ""

        self.no_options_label = Gtk.Label(halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER)
        self.no_options_label.set_line_wrap(True)
        self.no_options_label.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
        self.pack_end(self.no_options_label, True, True, 0)

    @property
    def filter(self) -> str:
        return self._filter

    @filter.setter
    def filter(self, value: str) -> None:
        """Sets the visibility of the options that have some text in the label or
        help-text."""
        self._filter = value
        self._filter_text = value.casefold()
        self.update_widgets()

    def generate_top_info_box(self, text):
        """Add a top section with general help text for the current tab"""
        help_box = Gtk.Box()
        help_box.set_margin_left(15)
        help_box.set_margin_right(15)
        help_box.set_margin_bottom(5)

        icon = Gtk.Image.new_from_icon_name("dialog-information", Gtk.IconSize.MENU)
        help_box.pack_start(icon, False, False, 5)

        title_label = Gtk.Label("<i>%s</i>" % text)
        title_label.set_line_wrap(True)
        title_label.set_alignment(0, 0.5)
        title_label.set_use_markup(True)
        help_box.pack_start(title_label, False, False, 5)

        self.pack_start(help_box, False, False, 0)
        self.pack_start(Gtk.HSeparator(), False, False, 12)

        help_box.show_all()

    def get_widget_generator(self) -> "ConfigWidgetGenerator":
        """Returns an object that creates option widgets and tracks them; this is
        lazy-initialized, but repeated calls return the same generator."""
        if self._widget_generator:
            return self._widget_generator

        gen = ConfigWidgetGenerator(self)

        if self.game and self.game.directory:
            gen.default_directory = self.game.directory
        elif self.game and self.game.has_runner:
            gen.default_directory = self.game.runner.working_dir
        elif self.lutris_config:
            gen.default_directory = self.lutris_config.system_config.get("game_path") or os.path.expanduser("~")
        else:
            gen.default_directory = os.path.expanduser("~")

        self._widget_generator = gen
        return gen

    def filter_widget(self, option_container: Gtk.Widget) -> bool:
        """Called by the widget generate to filter option containers; return true for
        those that should be visible."""
        if not self.advanced_visibility:
            is_advanced = hasattr(option_container, "lutris_advanced") and option_container.lutris_advanced
            if is_advanced:
                # Record that we hid this because it was advanced, not because of ordinary
                # visibility
                option_container.lutris_advanced_hidden = True  # type:ignore[attr-defined]
                return False

        filter_text = self._filter_text
        if filter_text and hasattr(option_container, "lutris_option_label"):
            label = option_container.lutris_option_label.casefold()
            helptext = option_container.lutris_option_helptext.casefold()  # type:ignore[attr-defined]
            if filter_text not in label and filter_text not in helptext:
                return False

        return True

    def generate_widgets(self):
        """Parse the config dict and generates widget accordingly."""
        # Select config section.
        if self.config_section == "game":
            self.config = self.lutris_config.game_config
            self.raw_config = self.lutris_config.raw_game_config
        elif self.config_section == "runner":
            self.config = self.lutris_config.runner_config
            self.raw_config = self.lutris_config.raw_runner_config
        elif self.config_section == "system":
            self.config = self.lutris_config.system_config
            self.raw_config = self.lutris_config.raw_system_config

        gen = self.get_widget_generator()

        # Go thru all options.
        for option in self.options:
            try:
                if "scope" in option:
                    if self.config_level not in option["scope"]:
                        continue

                # Generate option widget
                gen.add_container(option)
            except Exception as ex:
                logger.exception("Failed to generate option widget for '%s': %s", option.get("option"), ex)

        show_advanced = settings.read_setting("show_advanced_options") == "True"
        self._advanced_visibility = show_advanced
        gen.update_widgets()

    def update_widgets(self):
        if self._widget_generator:
            self._widget_generator.update_widgets()


class GameBox(ConfigBox):
    config_section = "game"

    def __init__(self, config_level: str, lutris_config: LutrisConfig, game: Game, **kwargs):
        ConfigBox.__init__(self, config_level, lutris_config, game, **kwargs)
        self.runner = game.runner
        if self.runner:
            self.options = self.runner.game_options
        else:
            logger.warning("No runner in game supplied to GameBox")


class RunnerBox(ConfigBox):
    """Configuration box for runner specific options"""

    config_section = "runner"

    def __init__(self, config_level: str, lutris_config: LutrisConfig, game: Game = None, **kwargs):
        ConfigBox.__init__(self, config_level, lutris_config, game, **kwargs)

        try:
            self.runner = import_runner(self.lutris_config.runner_slug)()
        except InvalidRunnerError:
            self.runner = None
        if self.runner:
            self.options = self.runner.get_runner_options()

        if lutris_config.level == "game":
            self.generate_top_info_box(
                _("If modified, these options supersede the same options from the base runner configuration.")
            )

    def generate_widgets(self):
        # Better safe than sorry - we search of Wine versions in directories
        # we do not control, so let's keep up to date more aggresively.
        clear_wine_version_cache()
        return super().generate_widgets()


class SystemConfigBox(ConfigBox):
    config_section = "system"

    def __init__(self, config_level: str, lutris_config: LutrisConfig, **kwargs) -> None:
        ConfigBox.__init__(self, config_level, lutris_config, **kwargs)
        self.runner = None
        runner_slug = self.lutris_config.runner_slug

        if runner_slug:
            self.options = sysoptions.with_runner_overrides(runner_slug)
        else:
            self.options = sysoptions.system_options

        if lutris_config.game_config_id and runner_slug:
            self.generate_top_info_box(
                _(
                    "If modified, these options supersede the same options from "
                    "the base runner configuration, which themselves supersede "
                    "the global preferences."
                )
            )
        elif runner_slug:
            self.generate_top_info_box(
                _("If modified, these options supersede the same options from the global preferences.")
            )


class ConfigWidgetGenerator(WidgetGenerator):
    def __init__(self, parent: ConfigBox) -> None:
        super().__init__(parent, parent.lutris_config)

        if parent.config is None or parent.raw_config is None:
            raise RuntimeError("Widgets can't be generated before the config is initialized.")

        self.config = parent.config
        self.raw_config = parent.raw_config
        self.lutris_config = parent.lutris_config
        self.reset_buttons: Dict[str, Gtk.Button] = {}

    def get_setting(self, option_key: str, default: Any) -> Any:
        if option_key in self.config:
            return self.config.get(option_key)
        else:
            return default

    def update_option_container(self, option, container: Gtk.Container, wrapper: Gtk.Container) -> None:
        super().update_option_container(option, container, wrapper)
        option_key = option["option"]

        if option_key in self.raw_config:
            set_option_wrapper_style_class(wrapper, "option-wrapper-assigned-here")
        else:
            default = self.get_default(option)
            value = self.get_setting(option_key, default)

            if value != default:
                set_option_wrapper_style_class(wrapper, "option-wrapper-non-default")
            else:
                set_option_wrapper_style_class(wrapper, None)

    def create_option_container(self, option: Dict[str, Any], wrapper: Gtk.Widget) -> Gtk.Container:
        option_key = option["option"]
        reset_container = Gtk.Box(visible=True)
        reset_container.set_margin_left(18)
        reset_container.pack_start(wrapper, True, True, 0)

        reset_button = Gtk.Button.new_from_icon_name("edit-undo-symbolic", Gtk.IconSize.MENU)
        reset_button.get_style_context().add_class("reset-button")
        reset_button.set_valign(Gtk.Align.CENTER)
        reset_button.set_halign(Gtk.Align.CENTER)
        reset_button.set_margin_bottom(6)
        reset_button.set_relief(Gtk.ReliefStyle.NONE)
        reset_button.set_tooltip_text(_("Reset option to global or default config"))
        reset_button.connect("clicked", self.on_reset_button_clicked, option)
        self.reset_buttons[option_key] = reset_button

        if option_key not in self.raw_config:
            reset_button.set_visible(False)
            reset_button.set_no_show_all(True)

        placeholder = Gtk.Box()
        placeholder.set_size_request(28, -1)

        placeholder.pack_start(reset_button, False, False, 0)
        reset_container.pack_end(placeholder, False, False, 5)
        return super().create_option_container(option, reset_container)

    def get_visibility(self, option: Dict[str, Any]) -> bool:
        option_container = self.option_containers[option["option"]]
        option_container.lutris_advanced_hidden = False  # type:ignore[attr-defined]
        option_visibility = super().get_visibility(option)

        if not option_visibility:
            return False

        return self.parent.filter_widget(option_container)

    def get_tooltip(self, option: Dict[str, Any], value: Any, default: Any):
        tooltip = super().get_tooltip(option, value, default)
        option_key = option["option"]
        if value != default and option_key not in self.raw_config:
            tooltip = tooltip + "\n\n" if tooltip else ""
            tooltip += _("<i>(Italic indicates that this option is modified in a lower configuration level.)</i>")
        return tooltip

    def update_widgets(self):
        super().update_widgets()

        def get_no_options_message() -> Optional[str]:
            if self.option_containers:
                for container in self.option_containers.values():
                    if container.get_visible():
                        return None

                if self.parent.filter:
                    return _("No options match '%s'") % self.parent.filter
                elif any(c.lutris_advanced_hidden for c in self.option_containers.values()):  # type:ignore[attr-defined]
                    return _("Only advanced options available")

            return _("No options available")

        message = get_no_options_message()
        if message:
            self.parent.no_options_label.set_text(message)
            self.parent.no_options_label.show()
        else:
            self.parent.no_options_label.hide()

    def on_reset_button_clicked(self, btn, option):
        """Clear option (remove from config, reset option widget)."""
        option_key = option["option"]
        wrapper = self.wrappers[option_key]

        btn.set_visible(False)

        self.raw_config.pop(option_key, None)
        self.lutris_config.update_cascaded_config()

        self.generate_widget(option, wrapper=wrapper)
        self.update_widgets()

    def on_changed(self, option_key, new_value):
        """Common actions when value changed on a widget"""
        self.raw_config[option_key] = new_value
        self.config[option_key] = new_value
        reset_btn = self.reset_buttons.get(option_key)

        if reset_btn:
            reset_btn.set_visible(True)

        super().on_changed(option_key, new_value)



"""
Discord Rich Presence Base Objects

THIS MODULE IS UNMAINTAINTED AND WILL BE MARKED FOR DEPRECATION UNLESS SOMEONE VOLUNTEERS TO PROPERLY MAINTAIN IT.
"""

from abc import ABCMeta, abstractmethod


class DiscordRichPresenceBase(metaclass=ABCMeta):
    """
    Discord Rich Presence Interface

    """

    @abstractmethod
    def update(self, discord_id: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    def clear(self) -> None:
        raise NotImplementedError()


class DiscordRPCNull(DiscordRichPresenceBase):
    """
    Null client for disabled Discord RPC
    """

    def update(self, discord_id: str) -> None:
        pass

    def clear(self) -> None:
        pass



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional

from lutris.util.strings import strip_accents

FLAG_TEXTS: Dict[str, Optional[bool]] = {"true": True, "yes": True, "false": False, "no": False}


def format_flag(flag: Optional[bool]) -> str:
    return "yes" if flag else "no"


class SearchPredicate(ABC):
    """This class is a filter function that also includes formatting and other functionality so
    it can be edited in the UI."""

    @abstractmethod
    def accept(self, candidate: Any) -> bool:
        """This method tests an object against the predicate."""
        return True

    def simplify(self) -> "SearchPredicate":
        """This method and return a simplified and flattened version of
        the predicate; the simplified predicate will match the same objects as
        the original."""
        return self

    def without_match(self, tag: str, value: Optional[str] = None) -> "SearchPredicate":
        """Returns a predicate without the MatchPredicate that has the tag and value
        given (or just the tag). Matches that are negated or the like are not removed."""
        return self

    def get_matches(self, tag: str) -> List[str]:
        """ "Returns all the values for the matches for the tag given; again any
        negated matches are ignored."""
        return []

    def without_flag(self, tag: str) -> "SearchPredicate":
        """Returns a predicate without the FlagPredicate that has the tag
        given. Flags that are negated or the like are not removed."""
        return self

    def has_flag(self, tag: str) -> bool:
        """True if the predicate has a FlagPredicate with the tag given;
        Flags that are negated or the like are ignored."""
        return False

    def get_flag(self, tag: str) -> Optional[bool]:
        """Returns the flag test value for the FlagPredicte with the tag
        given."""
        return None

    def to_child_text(self) -> str:
        """Returns the text of this predicate as it should be if nested inside a larger
        predicate; this may be in parentheses where __str__ would not be."""
        return str(self)

    @abstractmethod
    def __str__(self) -> str:
        pass


class FunctionPredicate(SearchPredicate):
    """This is a generate predicate that wraps a function to perform the test."""

    def __init__(self, predicate: Callable[[Any], bool], text: str) -> None:
        self.predicate = predicate
        self.text = text

    def accept(self, candidate: Any) -> bool:
        return self.predicate(candidate)

    def __str__(self):
        return self.text


class MatchPredicate(FunctionPredicate):
    """MatchPredicate is a predicate that tests a property against a value; you still provide
    a function to do the test, but the object records the tag and value explicitly for editing
    purposes."""

    def __init__(self, predicate: Callable[[Any], bool], text: str, tag: str, value: str) -> None:
        super().__init__(predicate, text)
        self.tag = tag
        self.value = value

    def get_matches(self, tag: str) -> List[str]:
        return [self.value] if self.tag == tag else []

    def without_match(self, tag: str, value: Optional[str] = None) -> "SearchPredicate":
        if value is None:
            return TRUE_PREDICATE if self.tag == tag else self

        return TRUE_PREDICATE if self.tag == tag and self.value == value else self


class FlagPredicate(SearchPredicate):
    """This is a predicate to match a boolean property. This odd setting is useful to override
    the default filtering Lutris provides, like filtering out hidden games."""

    def __init__(self, flag: Optional[bool], flag_function: Callable[[Any], bool], tag: str):
        self.flag = flag
        self.flag_function = flag_function
        self.tag = tag

    def accept(self, candidate: Any) -> bool:
        if self.flag is None:
            return True
        return self.flag == self.flag_function(candidate)

    def without_flag(self, tag: str) -> "SearchPredicate":
        return TRUE_PREDICATE if self.tag == tag else self

    def has_flag(self, tag: str) -> bool:
        return tag == self.tag

    def get_flag(self, tag: str) -> Optional[bool]:
        return self.flag if self.tag == tag else None

    def __str__(self):
        flag_text = format_flag(self.flag)
        return f"{self.tag}:{flag_text}"


class TextPredicate(SearchPredicate):
    """This is a predicate with no tag used to make text generically."""

    def __init__(self, match_text: str, text_function: Callable[[Any], Optional[str]], tag: str):
        self.tag = tag
        self.match_text = match_text
        self.stripped_text = strip_accents(match_text).casefold()
        self.text_function = text_function

    def accept(self, candidate: Any) -> bool:
        candidate_text = self.text_function(candidate)
        if not candidate_text:
            return False

        candidate_text = strip_accents(candidate_text).casefold()
        return bool(candidate_text and self.stripped_text in candidate_text)

    def __str__(self):
        if self.tag:
            return f"{self.tag}:{self.match_text}"
        return self.match_text


class NotPredicate(SearchPredicate):
    """This predicate reverses the effect of another, and also 'hides' it from
    editing methods."""

    def __init__(self, to_negate: SearchPredicate) -> None:
        self.to_negate = to_negate

    def accept(self, candidate: Any) -> bool:
        return not self.to_negate.accept(candidate)

    def to_child_text(self) -> str:
        return f"(-{self.to_negate.to_child_text()})"

    def __str__(self):
        return "-" + str(self.to_negate)


class AndPredicate(SearchPredicate):
    """This predicate combines other predicates so all must accept a candidate
    before it is accepted."""

    def __init__(self, components: List[SearchPredicate]) -> None:
        self.components = components

    def accept(self, candidate: Any) -> bool:
        for c in self.components:
            if not c.accept(candidate):
                return False
        return True

    def simplify(self) -> "SearchPredicate":
        simplified = []
        for c in self.components:
            c = c.simplify()
            if c != TRUE_PREDICATE:
                if isinstance(c, AndPredicate):
                    simplified += c.components
                else:
                    simplified.append(c)
        return AndPredicate(simplified) if simplified else TRUE_PREDICATE

    def get_matches(self, tag: str) -> List[str]:
        matches = []
        for c in self.components:
            matches += c.get_matches(tag)
        return matches

    def without_match(self, tag: str, value: Optional[str] = None) -> "SearchPredicate":
        new_components = []
        for c in self.components:
            r = c.without_match(tag, value)
            if r != TRUE_PREDICATE:
                new_components.append(r)

        if new_components != self.components:
            return AndPredicate(new_components)

        return self

    def without_flag(self, tag: str) -> "SearchPredicate":
        new_components = []
        for c in self.components:
            r = c.without_flag(tag)
            if r != TRUE_PREDICATE:
                new_components.append(r)

        if new_components != self.components:
            return AndPredicate(new_components)

        return self

    def has_flag(self, tag: str) -> bool:
        for c in self.components:
            if c.has_flag(tag):
                return True
        return False

    def get_flag(self, tag: str) -> Optional[bool]:
        for c in self.components:
            if c.has_flag(tag):
                return c.get_flag(tag)
        return None

    def __str__(self):
        return " ".join(str(c) for c in self.components)


class OrPredicate(SearchPredicate):
    """This predicate combines other predicates so that a candidate is accepted
    if any component accepts it. This also hides the components from editing."""

    def __init__(self, components: List[SearchPredicate]) -> None:
        self.components = components

    def accept(self, candidate: Any) -> bool:
        for c in self.components:
            if c.accept(candidate):
                return True
        return False

    def simplify(self) -> "SearchPredicate":
        simplified = []
        for c in self.components:
            c = c.simplify()
            if c == TRUE_PREDICATE:
                return c
            if isinstance(c, OrPredicate):
                simplified += c.components
            else:
                simplified.append(c)
        return OrPredicate(simplified)

    def to_child_text(self) -> str:
        return f"({self})"

    def __str__(self):
        return " OR ".join(c.to_child_text() for c in self.components)


class TruePredicate(SearchPredicate):
    """This predicate matches everything."""

    def accept(self, candidate: Any) -> bool:
        return True

    def __str__(self):
        return ""


TRUE_PREDICATE: SearchPredicate = TruePredicate()



"""Automatically detects game executables in a folder"""

import os

from lutris.util import magic, system
from lutris.util.log import logger


def is_excluded_elf(filename):
    excluded = ("xdg-open", "uninstall")
    _fn = filename.lower()
    return any(exclude in _fn for exclude in excluded)


def find_linux_game_executable(path, make_executable=False):
    """Looks for a binary or shell script that launches the game in a directory"""

    for base, _dirs, files in os.walk(path):
        candidates = {}
        for _file in files:
            if is_excluded_elf(_file):
                continue
            abspath = os.path.join(base, _file)
            file_type = magic.from_file(abspath)
            if "ASCII text executable" in file_type:
                candidates["shell"] = abspath
            if "Bourne-Again shell script" in file_type:
                candidates["bash"] = abspath
            if "POSIX shell script executable" in file_type:
                candidates["posix"] = abspath
            if "64-bit LSB executable" in file_type or "64-bit LSB shared object" in file_type:
                candidates["64bit"] = abspath
            if "32-bit LSB executable" in file_type or "32-bit LSB shared object" in file_type:
                candidates["32bit"] = abspath
        if candidates:
            if make_executable:
                for candidate in candidates.values():
                    system.make_executable(candidate)
            return (
                candidates.get("shell")
                or candidates.get("bash")
                or candidates.get("posix")
                or candidates.get("64bit")
                or candidates.get("32bit")
            )
    logger.error("Couldn't find a Linux executable in %s", path)
    return ""


def is_excluded_dir(path):
    excluded = (
        "Internet Explorer",
        "Windows NT",
        "Common Files",
        "Windows Media Player",
        "windows",
        "ProgramData",
        "users",
        "GameSpy Arcade",
    )
    return any(dir_name in excluded for dir_name in path.split("/"))


def is_excluded_exe(filename):
    excluded = (
        "unins000",
        "uninstal",
        "update",
        "setup",
        "config.exe",
        "gsarcade.exe",
        "dosbox.exe",
    )
    _fn = filename.lower()
    return any(exclude in _fn for exclude in excluded)


def find_windows_game_executable(path):
    for base, _dirs, files in os.walk(path):
        candidates = {}
        if is_excluded_dir(base):
            continue
        for _file in files:
            if is_excluded_exe(_file):
                continue
            abspath = os.path.join(base, _file)
            if os.path.islink(abspath):
                continue
            file_type = magic.from_file(abspath)
            if "MS Windows shortcut" in file_type:
                candidates["link"] = abspath
            elif all([_ in file_type for _ in ("PE32+ executable", "(GUI)", "x86-64")]):
                candidates["64bit"] = abspath
            elif all([_ in file_type for _ in ("PE32 executable", "(GUI)")]) and any(
                [_ in file_type for _ in ("Intel i386", "Intel 80386")]
            ):
                candidates["32bit"] = abspath
        if candidates:
            return candidates.get("link") or candidates.get("64bit") or candidates.get("32bit")
    logger.error("Couldn't find a Windows executable in %s", path)
    return ""



"""Handle the game, runner and global system configurations."""

import os
import time
from shutil import copyfile
from typing import Any, Dict, Optional, Set

from lutris import settings, sysoptions
from lutris.runners import InvalidRunnerError, import_runner
from lutris.util.log import logger
from lutris.util.system import path_exists
from lutris.util.yaml import read_yaml_from_file, write_yaml_to_file


def make_game_config_id(game_slug: str) -> str:
    """Return an unique config id to avoid clashes between multiple games"""
    return "{}-{}".format(game_slug, int(time.time()))


def write_game_config(game_slug: str, config: Dict[str, Any]) -> str:
    """Writes a game config to disk"""
    configpath = make_game_config_id(game_slug)
    logger.debug("Writing game config to %s", configpath)
    config_filename = os.path.join(settings.CONFIG_DIR, "games/%s.yml" % configpath)
    write_yaml_to_file(config, config_filename)
    return configpath


def duplicate_game_config(game_slug: str, source_config_id: str) -> str:
    """Copies an existing configuration file, giving it a new id that this
    function returns."""
    new_config_id = make_game_config_id(game_slug)
    src_path = os.path.join(settings.CONFIG_DIR, "games/%s.yml" % source_config_id)
    dest_path = os.path.join(settings.CONFIG_DIR, "games/%s.yml" % new_config_id)
    copyfile(src_path, dest_path)
    return new_config_id


def rename_config(old_config_id: str, new_slug: str) -> Optional[str]:
    old_slug, timestamp = old_config_id.rsplit("-", 1)
    if old_slug == new_slug:
        return None
    new_config_id = f"{new_slug}-{timestamp}"
    src_path = f"{settings.GAME_CONFIG_DIR}/{old_config_id}.yml"
    dest_path = f"{settings.GAME_CONFIG_DIR}/{new_config_id}.yml"
    os.rename(src_path, dest_path)
    return new_config_id


class LutrisConfig:
    """Class where all the configuration handling happens.

    Description
    ===========
    Lutris' configuration uses a cascading mechanism where
    each higher, more specific level overrides the lower ones

    The levels are (highest to lowest): `game`, `runner` and `system`.
    Each level has its own set of options (config section), available to and
    overridden by upper levels:
    ```
     level | Config sections
    -------|----------------------
      game | system, runner, game
    runner | system, runner
    system | system
    ```
    Example: if requesting runner options at game level, their returned value
    will be from the game level config if it's set at this level; if not it
    will be the value from runner level if available; and if not, the default
    value set in the runner's module, or None.

    The config levels are stored in separate YAML format text files.

    Usage
    =====
    The config level will be auto set depending on what you pass to __init__:
    - For game level, pass game_config_id and optionally runner_slug (better perfs)
    - For runner level, pass runner_slug
    - For system level, pass nothing
    If need be, you can pass the level manually.

    To read, use the config sections dicts: game_config, runner_config and
    system_config.

    To write, modify the relevant `raw_*_config` section dict, then run
    `save()`.

    """

    def __init__(
        self,
        runner_slug: Optional[str] = None,
        game_config_id: Optional[str] = None,
        level: Optional[str] = None,
        options_supported: Optional[Set[str]] = None,
    ):
        self.game_config_id = game_config_id
        if runner_slug:
            self.runner_slug: Optional[str] = str(runner_slug)
        else:
            self.runner_slug: Optional[str] = runner_slug

        self.options_supported = options_supported
        # Cascaded config sections (for reading)
        self.game_config = {}
        self.runner_config = {}
        self.system_config = {}

        # Raw (non-cascaded) sections (for writing)
        self.raw_game_config = {}
        self.raw_runner_config = {}
        self.raw_system_config = {}

        self.raw_config = {}

        # Set config level
        self.level = level
        if not level:
            if game_config_id:
                self.level = "game"
            elif runner_slug:
                self.level = "runner"
            else:
                self.level = "system"
        self.initialize_config()

    def __repr__(self):
        return "LutrisConfig(level=%s, game_config_id=%s, runner=%s)" % (
            self.level,
            self.game_config_id,
            self.runner_slug,
        )

    @property
    def system_config_path(self) -> str:
        return os.path.join(settings.CONFIG_DIR, "system.yml")

    @property
    def runner_config_path(self) -> Optional[str]:
        if not self.runner_slug:
            return None
        return os.path.join(settings.RUNNERS_CONFIG_DIR, "%s.yml" % self.runner_slug)

    @property
    def game_config_path(self) -> Optional[str]:
        if not self.game_config_id:
            return None
        return os.path.join(settings.CONFIG_DIR, "games/%s.yml" % self.game_config_id)

    def initialize_config(self) -> None:
        """Init and load config files"""
        self.game_level = {"system": {}, self.runner_slug: {}, "game": {}}
        self.runner_level = {"system": {}, self.runner_slug: {}}
        self.system_level = {"system": {}}
        if self.game_config_path:
            self.game_level.update(read_yaml_from_file(self.game_config_path))
        if self.runner_config_path:
            self.runner_level.update(read_yaml_from_file(self.runner_config_path))
        self.system_level.update(read_yaml_from_file(self.system_config_path))

        self.update_cascaded_config()
        self.update_raw_config()

    def update_cascaded_config(self) -> None:
        if self.system_level.get("system") is None:
            self.system_level["system"] = {}
        self.system_config.clear()
        self.system_config.update(self.get_defaults("system"))
        self.system_config.update(self.system_level.get("system", {}))

        if self.level in ["runner", "game"] and self.runner_slug:
            if self.runner_level.get(self.runner_slug) is None:
                self.runner_level[self.runner_slug] = {}
            if self.runner_level.get("system") is None:
                self.runner_level["system"] = {}
            self.runner_config.clear()
            self.runner_config.update(self.get_defaults("runner"))
            self.runner_config.update(self.runner_level.get(self.runner_slug, {}))
            self.merge_to_system_config(self.runner_level.get("system"))

        if self.level == "game" and self.runner_slug:
            if self.game_level.get("game") is None:
                self.game_level["game"] = {}
            if self.game_level.get(self.runner_slug) is None:
                self.game_level[self.runner_slug] = {}
            if self.game_level.get("system") is None:
                self.game_level["system"] = {}
            self.game_config.clear()
            self.game_config.update(self.get_defaults("game"))
            self.game_config.update(self.game_level.get("game", {}))
            self.runner_config.update(self.game_level.get(self.runner_slug, {}))
            self.merge_to_system_config(self.game_level.get("system"))

    def merge_to_system_config(self, config: Optional[Dict[str, Any]]) -> None:
        """Merge a configuration to the system configuration"""
        if config:
            existing_env = None
            if self.system_config.get("env") and "env" in config:
                existing_env = self.system_config["env"]
            self.system_config.update(config)
            if existing_env:
                self.system_config["env"] = existing_env
                self.system_config["env"].update(config["env"])

        # Don't save env items where the key is empty; this would crash when used.
        if "env" in self.system_config:
            self.system_config["env"] = {k: v for k, v in self.system_config["env"].items() if k}

    def update_raw_config(self) -> None:
        # Select the right level of config
        if self.level == "game":
            raw_config = self.game_level
        elif self.level == "runner":
            raw_config = self.runner_level
        else:
            raw_config = self.system_level

        # Load config sections
        self.raw_system_config = raw_config["system"]
        if self.level in ["runner", "game"] and self.runner_slug is not None:
            self.raw_runner_config = raw_config[self.runner_slug]
        if self.level == "game":
            self.raw_game_config = raw_config["game"]

        self.raw_config = raw_config

    def remove(self) -> None:
        """Delete the configuration file from disk."""
        if not self.game_config_path or not path_exists(self.game_config_path):
            logger.debug("No config file at %s", self.game_config_path)
            return
        os.remove(self.game_config_path)
        logger.debug("Removed config %s", self.game_config_path)

    def save(self) -> None:
        """Save configuration file according to its type"""

        if self.options_supported is not None:
            raise RuntimeError("LutrisConfig instances that are restricted to only some options can't be saved.")

        if self.level == "system":
            config = self.system_level
            config_path: str = self.system_config_path
        elif self.level == "runner" and self.runner_config_path:
            config = self.runner_level
            config_path = self.runner_config_path
        elif self.level == "game" and self.game_config_path:
            config = self.game_level
            config_path = self.game_config_path
        else:
            raise ValueError("Invalid config level '%s'" % self.level)
        # Remove keys with no values from config before saving
        config = {key: value for key, value in config.items() if value}
        logger.debug("Saving %s config to %s", self, config_path)
        write_yaml_to_file(config, config_path)
        self.initialize_config()

    def get_defaults(self, options_type: str) -> Dict[str, Any]:
        """Return a dict of options' default value."""
        options_dict = self.options_as_dict(options_type)
        defaults = {}
        for option, params in options_dict.items():
            if "default" in params:
                default = params["default"]
                if callable(default):
                    if self.options_supported is None or option in self.options_supported:
                        try:
                            default = default()
                        except Exception as ex:
                            logger.exception("Unable to generate a default for '%s': %s", option, ex)
                            continue
                    else:
                        # Do not evaluate options we aren't supposed to use, in case
                        # this is expensive or unsafe.
                        default = None
                defaults[option] = default
        return defaults

    def options_as_dict(self, options_type: str) -> Dict[str, Any]:
        """Convert the option list to a dict with option name as keys"""
        if options_type == "system":
            options = (
                sysoptions.with_runner_overrides(self.runner_slug) if self.runner_slug else sysoptions.system_options
            )
        else:
            if not self.runner_slug:
                return {}
            attribute_name = options_type + "_options"

            try:
                runner = import_runner(self.runner_slug)
            except InvalidRunnerError:
                options = {}
            else:
                if not getattr(runner, attribute_name):
                    runner = runner()

                options = getattr(runner, attribute_name)
        return dict((opt["option"], opt) for opt in options)



import time
from typing import List

from gi.repository import Gdk, GObject, Gtk

from lutris.database.games import get_game_for_service
from lutris.database.services import ServiceGameCollection
from lutris.game import GAME_START, Game
from lutris.game_actions import GameActions, get_game_actions
from lutris.gui.widgets import EMPTY_NOTIFICATION_REGISTRATION
from lutris.gui.widgets.contextual_menu import ContextualMenu
from lutris.gui.widgets.utils import MEDIA_CACHE_INVALIDATED, get_application
from lutris.util.jobs import schedule_repeating_at_idle
from lutris.util.log import logger
from lutris.util.path_cache import MISSING_GAMES


class GameView:
    # pylint: disable=no-member
    __gsignals__ = {
        "game-selected": (GObject.SIGNAL_RUN_FIRST, None, (object,)),
        "game-activated": (GObject.SIGNAL_RUN_FIRST, None, (str,)),
    }

    def __init__(self):
        self.game_store = None
        self.service = None
        self.service_media = None
        self.cache_notification_registration = EMPTY_NOTIFICATION_REGISTRATION
        self.missing_games_updated_registration = EMPTY_NOTIFICATION_REGISTRATION
        self.game_start_registration = EMPTY_NOTIFICATION_REGISTRATION
        self.image_renderer = None

    def connect_signals(self):
        """Signal handlers common to all views"""
        self.cache_notification_registration = MEDIA_CACHE_INVALIDATED.register(self.on_media_cache_invalidated)
        self.missing_games_updated_registration = MISSING_GAMES.updated.register(self.on_missing_games_updated)

        self.connect("destroy", self.on_destroy)
        self.connect("button-press-event", self.popup_contextual_menu)
        self.connect("key-press-event", self.handle_key_press)

        self.game_start_registration = GAME_START.register(self.on_game_start)

    def set_game_store(self, game_store):
        self.game_store = game_store
        self.service = game_store.service
        self.service_media = game_store.service_media

        if self.image_renderer:
            self.image_renderer.service = self.service

    def on_media_cache_invalidated(self):
        self.queue_draw()

    def on_missing_games_updated(self):
        if self.image_renderer and self.image_renderer.show_badges:
            self.queue_draw()

    def on_destroy(self, _widget) -> None:
        self.cache_notification_registration.unregister()
        self.missing_games_updated_registration.unregister()
        self.game_start_registration.unregister()

    def popup_contextual_menu(self, view, event):
        """Contextual menu."""
        if event.button != Gdk.BUTTON_SECONDARY:
            return
        current_path = self.get_path_at(event.x, event.y)
        if current_path:
            selection = self.get_selected()
            if current_path not in selection:
                selection = [current_path]
                self.set_selected(selection)

            game_actions = self.get_game_actions_for_paths(selection)
            contextual_menu = ContextualMenu(game_actions.get_game_actions())
            contextual_menu.popup(event, game_actions)
            return True

    def get_selected_game_actions(self) -> GameActions:
        return self.get_game_actions_for_paths(self.get_selected())

    def get_game_actions_for_paths(self, paths) -> GameActions:
        game_ids = []
        for path in paths:
            game_ids.append(self.get_game_id_for_path(path))
        games = self._get_games_by_ids(game_ids)
        return get_game_actions(games, window=self.get_toplevel())

    def _get_games_by_ids(self, game_ids: List[str]) -> List[Game]:
        """Resolves a list of game-ids to a list of game objects,
        looking up running games, service games and all that."""

        def _get_game_by_id(id_to_find: str) -> Game:
            application = get_application()
            return application.get_game_by_id(id_to_find) if application else Game(id_to_find)

        games = []
        for game_id in game_ids:
            if self.service:
                db_game = get_game_for_service(self.service.id, game_id)

                if db_game:
                    if db_game["id"]:
                        games.append(_get_game_by_id(db_game["id"]))
                else:
                    db_game = ServiceGameCollection.get_game(self.service.id, game_id)
                    games.append(Game.create_empty_service_game(db_game, self.service))
            elif game_id:
                games.append(_get_game_by_id(game_id))

        return games

    def get_selected_game_id(self):
        """Returns the ID of the selected game, if there is exactly one- or
        None if there is no selection or a multiple-selection."""
        selected = self.get_selected()
        if len(selected) == 1:
            return self.get_game_id_for_path(selected[0])
        return None

    def handle_key_press(self, widget, event):  # pylint: disable=unused-argument
        try:
            key = event.keyval
            if key == Gdk.KEY_Delete:
                game_actions = self.get_selected_game_actions()
                if game_actions.is_game_removable:
                    game_actions.on_remove_game(self)
            elif key == Gdk.KEY_Break:
                game_actions = self.get_selected_game_actions()
                if game_actions.is_game_running:
                    game_actions.on_game_stop(self)
        except Exception as ex:
            logger.exception("Unable to handle key press: %s", ex)

    def get_toplevel(self):
        raise NotImplementedError()

    def get_selected(self):
        return []

    def set_selected(self, paths, scroll_into_view=False):
        raise NotImplementedError()

    def get_game_id_for_path(self, path):
        raise NotImplementedError()

    def get_path_for_game_id(self, game_id):
        raise NotImplementedError()

    def on_game_start(self, game: Game) -> None:
        """On game start, we trigger an animation to show the game is starting; it runs at least
        one cycle, but continues until the game exits the STATE_LAUNCHING state."""

        # We animate by looking at how long the animation has been running;
        # This keeps things on track even if drawing is low or the timeout we use
        # is not quite regular.

        start_time = time.monotonic()
        cycle_time = 0.375
        max_indent = 0.1
        toplevel = self.get_toplevel()
        paused = False

        def is_modally_blocked():
            # Is there a modal dialog that is blocking our top-level parent?
            # if so we want to pause the animation.
            for w in Gtk.Window.list_toplevels():
                if w != toplevel and isinstance(w, Gtk.Dialog):
                    if w.get_modal() and w.get_transient_for() == toplevel:
                        return True

        def animate():
            nonlocal paused, start_time

            now = time.monotonic()
            elapsed = now - start_time

            if elapsed > cycle_time:
                # Check for stopping and pausing only at cycle end, so we don't do it too often,
                # and to avoid a janky looking visible snap-back to full size.
                if game.state != game.STATE_LAUNCHING:
                    if self.image_renderer.inset_game(game.id, 0.0):
                        self.queue_draw()
                    return False

                start_time = now
                paused = is_modally_blocked()

            cycle = elapsed % cycle_time

            # After 1/2 the cycle, start counting down instead of up
            if cycle > cycle_time / 2:
                cycle = cycle_time - cycle

            # scale to achieve the max_indent at cycle_time/2.
            if paused:
                fraction = 0.0
            else:
                fraction = max_indent * (cycle * 2 / cycle_time)

            if self.image_renderer.inset_game(game.id, fraction):
                self.queue_draw()

            return True  # Return True to call again after another timeout

        if self.image_renderer:
            schedule_repeating_at_idle(animate, interval_seconds=0.025)



# Standard Library
from mmap import mmap


def scan_to_00(mm, start):
    """Read bytes from the mm mmap, beggining at the start
    offset and ending at the first 0x00.

    Return:
        bytes
    """
    buff = b""
    achar = None
    number = start
    while achar != 0:
        achar = mm[number]
        if achar != 0:
            buff += bytes((achar,))
        number += 1
    return buff


def bytes_to_str(byte):
    """transform bytes to string with the default codec"""
    return str(byte)[2:-1]


def rom_read_data(location):
    """extract data from the rom location at location.
    return a dict with "data" and "config", to be applied to a game in Lutris"""
    # TODO: extract the image of the rom
    data = {}
    with open(location, "r+", encoding="utf-8") as rom:
        mm = mmap(rom.fileno(), 0)
        # the most of the scan of the game
        if mm[0:4] == b"WBFS":  # wii WBFS file
            data["name"] = bytes_to_str(scan_to_00(mm, 0x220))
            data["slug"] = "wii-" + bytes_to_str(scan_to_00(mm, 0x200))
        elif mm[0x18:0x1C] == b"\x5d\x1c\x9e\xa3":  # wii iso file
            data["name"] = bytes_to_str(scan_to_00(mm, 0x20))
            data["slug"] = "wii-" + bytes_to_str(scan_to_00(mm, 0x0))
        else:
            return False
    return data



from sqlite3 import OperationalError

from lutris.database.games import get_games
from lutris.game import Game
from lutris.util.log import logger


def migrate():
    """Put all previously hidden games into the new '.hidden' category."""
    logger.info("Moving hidden games to the '.hidden' category")
    try:
        game_ids = [g["id"] for g in get_games(filters={"hidden": 1})]
    except OperationalError:
        # A brand-new DB will not have the hidden column at all,
        # so no migration is required.
        return

    for game_id in game_ids:
        game = Game(game_id)
        game.mark_as_hidden(True)
        logger.info("Migrated '%s' to '.hidden' category.", game.name)



import os
from gettext import gettext as _
from gettext import ngettext

from gi.repository import Gio, Gtk

from lutris import api, sysoptions
from lutris.config import LutrisConfig
from lutris.gui.config.add_game_dialog import AddGameDialog
from lutris.gui.dialogs import ErrorDialog, ModelessDialog
from lutris.gui.dialogs.game_import import ImportGameDialog
from lutris.gui.widgets.common import FileChooserEntry
from lutris.gui.widgets.navigation_stack import NavigationStack
from lutris.installer import AUTO_WIN32_EXE, get_installers
from lutris.util.jobs import COMPLETED_IDLE_TASK, AsyncCall, schedule_at_idle
from lutris.util.strings import gtk_safe, slugify
from lutris.util.wine.proton import is_proton_version
from lutris.util.wine.wine import GE_PROTON_LATEST


class AddGamesWindow(ModelessDialog):  # pylint: disable=too-many-public-methods
    """Show a selection of ways to add games to Lutris"""

    sections = [
        (
            "system-search-symbolic",
            "go-next-symbolic",
            _("Search the Lutris website for installers"),
            _("Query our website for community installers"),
            "search_installers",
        ),
        (
            "application-x-executable-symbolic",
            "go-next-symbolic",
            _("Install a Windows game from an executable"),
            _("Launch a Windows executable (.exe) installer"),
            "install_from_setup",
        ),
        (
            "x-office-document-symbolic",
            "go-next-symbolic",
            _("Install from a local install script"),
            _("Run a YAML install script"),
            "install_from_script",
        ),
        (
            "application-x-firmware-symbolic",
            "go-next-symbolic",
            _("Import ROMs"),
            _("Import ROMs referenced in TOSEC, No-intro or Redump"),
            "import_rom",
        ),
        (
            "list-add-symbolic",
            "view-more-horizontal-symbolic",
            _("Add locally installed game"),
            _("Manually configure a game available locally"),
            "add_local_game",
        ),
    ]

    def __init__(self, **kwargs):
        ModelessDialog.__init__(self, title=_("Add games to Lutris"), use_header_bar=True, **kwargs)
        self.set_default_size(640, 450)
        self.search_entry = None
        self.search_frame = None
        self.search_explanation_label = None
        self.search_listbox = None
        self.search_timer_task = COMPLETED_IDLE_TASK
        self.search_spinner = None
        self.text_query = None
        self.search_result_label = None

        content_area = self.get_content_area()

        self.page_title_label = Gtk.Label(visible=True)
        content_area.pack_start(self.page_title_label, False, False, 0)

        self.accelerators = Gtk.AccelGroup()
        self.add_accel_group(self.accelerators)

        header_bar = self.get_header_bar()

        self.back_button = Gtk.Button(_("Back"), no_show_all=True)
        self.back_button.connect("clicked", self.on_back_clicked)
        key, mod = Gtk.accelerator_parse("<Alt>Left")
        self.back_button.add_accelerator("clicked", self.accelerators, key, mod, Gtk.AccelFlags.VISIBLE)
        key, mod = Gtk.accelerator_parse("<Alt>Home")
        self.accelerators.connect(key, mod, Gtk.AccelFlags.VISIBLE, self.on_navigate_home)
        header_bar.pack_start(self.back_button)

        self.continue_button = Gtk.Button(_("_Continue"), no_show_all=True, use_underline=True)
        header_bar.pack_end(self.continue_button)
        self.continue_handler = None

        self.cancel_button = Gtk.Button(_("Cancel"), use_underline=True)
        self.cancel_button.connect("clicked", self.on_cancel_clicked)
        key, mod = Gtk.accelerator_parse("Escape")
        self.accelerators.connect(key, mod, Gtk.AccelFlags.VISIBLE, lambda *_args: self.destroy())
        header_bar.pack_start(self.cancel_button)
        header_bar.set_show_close_button(False)

        content_area.set_margin_top(18)
        content_area.set_margin_bottom(18)
        content_area.set_margin_right(18)
        content_area.set_margin_left(18)
        content_area.set_spacing(12)

        self.stack = NavigationStack(self.back_button, cancel_button=self.cancel_button)
        content_area.pack_start(self.stack, True, True, 0)

        # Pre-create some controls so they can be used in signal handlers

        self.scan_directory_chooser = FileChooserEntry(
            title=_("Select folder"), action=Gtk.FileChooserAction.SELECT_FOLDER
        )

        self.install_from_setup_game_name_entry = Gtk.Entry()
        self.install_from_setup_game_slug_checkbox = Gtk.CheckButton(label=_("Identifier"))
        self.install_from_setup_game_slug_entry = Gtk.Entry(sensitive=False)
        self.install_from_setup_game_slug_entry.connect(
            "focus-out-event", self.on_install_from_setup_game_slug_entry_focus_out
        )
        self.installer_presets = Gtk.ListStore(str, str)
        self.install_preset_dropdown = Gtk.ComboBox.new_with_model(self.installer_presets)
        self.installer_locale = Gtk.ListStore(str, str)
        self.install_locale_dropdown = Gtk.ComboBox.new_with_model(self.installer_locale)

        self.install_script_file_chooser = FileChooserEntry(title=_("Select script"), action=Gtk.FileChooserAction.OPEN)

        self.import_rom_file_chooser = FileChooserEntry(
            title=_("Select ROMs"), action=Gtk.FileChooserAction.SELECT_FOLDER
        )

        self.stack.add_named_factory("initial", self.create_initial_page)
        self.stack.add_named_factory("search_installers", self.create_search_installers_page)
        self.stack.add_named_factory("install_from_setup", self.create_install_from_setup_page)
        self.stack.add_named_factory("install_from_script", self.create_install_from_script_page)
        self.stack.add_named_factory("import_rom", self.create_import_rom_page)

        self.show_all()

        self.load_initial_page()

    def on_back_clicked(self, _widget):
        self.stack.navigate_back()

    def on_navigate_home(self, _accel_group, _window, _keyval, _modifier):
        self.stack.navigate_home()

    def on_cancel_clicked(self, _widget):
        self.destroy()

    # Initial Page

    def load_initial_page(self):
        self.stack.navigate_to_page(self.present_inital_page)

    def create_initial_page(self):
        frame = Gtk.Frame(shadow_type=Gtk.ShadowType.ETCHED_IN)
        listbox = Gtk.ListBox()
        listbox.set_activate_on_single_click(True)
        for icon, next_icon, text, subtext, callback_name in self.sections:
            row = self._get_listbox_row(icon, text, subtext, next_icon)
            row.callback_name = callback_name

            listbox.add(row)
        listbox.connect("row-activated", self.on_row_activated)
        frame.add(listbox)
        return frame

    def present_inital_page(self):
        self.set_page_title_markup(None)
        self.stack.present_page("initial")
        self.display_cancel_button()

    def on_row_activated(self, listbox, row):
        if row.callback_name:
            callback = getattr(self, row.callback_name)
            callback()

    # Search Installers Page

    def search_installers(self):
        """Search installers with the Lutris API"""
        if self.search_entry:
            self.search_entry.set_text("")
            self.search_result_label.set_text("")
            self.search_result_label.hide()
            self.search_frame.hide()
            self.search_explanation_label.show()
        self.stack.navigate_to_page(self.present_search_installers_page)

    def create_search_installers_page(self):
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, no_show_all=True, spacing=6, visible=True)
        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, visible=True)
        self.search_entry = Gtk.SearchEntry(visible=True)
        hbox.pack_start(self.search_entry, True, True, 0)
        self.search_spinner = Gtk.Spinner(visible=False)
        hbox.pack_end(self.search_spinner, False, False, 6)
        vbox.pack_start(hbox, False, False, 0)
        self.search_result_label = self._get_label("")
        self.search_result_label.hide()
        vbox.pack_start(self.search_result_label, False, False, 0)
        self.search_entry.connect("changed", self._on_search_updated)

        explanation = _(
            "Lutris will search Lutris.net for games matching the terms you enter, and any "
            "that it finds will appear here.\n\n"
            "When you click on a game that it found, the installer window will appear to "
            "perform the installation."
        )

        self.search_explanation_label = self._get_explanation_label(explanation)
        vbox.add(self.search_explanation_label)

        self.search_frame = Gtk.Frame(shadow_type=Gtk.ShadowType.ETCHED_IN)
        self.search_listbox = Gtk.ListBox(visible=True)
        self.search_listbox.connect("row-activated", self._on_game_selected)
        scroll = Gtk.ScrolledWindow(visible=True)
        scroll.set_vexpand(True)
        scroll.add(self.search_listbox)
        self.search_frame.add(scroll)

        vbox.pack_start(self.search_frame, True, True, 0)
        return vbox

    def present_search_installers_page(self):
        self.set_page_title_markup(_("<b>Search Lutris.net</b>"))
        self.stack.present_page("search_installers")
        self.search_entry.grab_focus()
        self.display_cancel_button()

    def _on_search_updated(self, entry):
        self.search_timer_task.unschedule()
        self.text_query = entry.get_text().strip()
        self.search_timer_task = schedule_at_idle(self.update_search_results, delay_seconds=0.75)

    def update_search_results(self) -> None:
        # Don't start a search while another is going; defer it instead.
        if self.search_spinner.get_visible():
            self.search_timer_task = schedule_at_idle(self.update_search_results, delay_seconds=0.75)
            return

        if self.text_query:
            self.search_spinner.show()
            self.search_spinner.start()
            AsyncCall(api.search_games, self.update_search_results_cb, self.text_query)

    def update_search_results_cb(self, api_games, error):
        if error:
            raise error

        self.search_spinner.stop()
        self.search_spinner.hide()
        total_count = api_games.get("count", 0)
        count = len(api_games.get("results", []))

        if not count:
            self.search_result_label.set_markup(_("No results"))
        elif count == total_count:
            text = ngettext("Showing <b>%d</b> result", "Showing <b>%d</b> results", count) % count
            self.search_result_label.set_markup(text)
        else:
            self.search_result_label.set_markup(_("<b>%s</b> results, only displaying first %s") % (total_count, count))
        for row in self.search_listbox.get_children():
            row.destroy()
        for game in api_games.get("results", []):
            platforms = ",".join(gtk_safe(platform["name"]) for platform in game["platforms"])
            year = game["year"] or ""
            if platforms and year:
                platforms = ", " + platforms

            row = self._get_listbox_row("", gtk_safe(game["name"]), f"{year}{platforms}")
            row.api_info = game
            self.search_listbox.add(row)
        self.search_result_label.show()
        self.search_frame.show()
        self.search_explanation_label.hide()

    def _on_game_selected(self, listbox, row):
        game_slug = row.api_info["slug"]
        application = Gio.Application.get_default()
        application.show_lutris_installer_window(game_slug=game_slug)
        self.destroy()

    # Install from Setup Page

    def install_from_setup(self):
        """Install from a setup file"""
        self.stack.navigate_to_page(self.present_install_from_setup_page)

    def create_install_from_setup_page(self):
        name_label = self._get_label(_("Game name"))

        self.install_from_setup_game_name_entry.set_hexpand(True)
        self.install_from_setup_game_slug_entry.set_hexpand(True)

        grid = Gtk.Grid(row_spacing=6, column_spacing=6)
        grid.set_column_homogeneous(False)
        grid.attach(name_label, 0, 0, 1, 1)
        grid.attach(self.install_from_setup_game_name_entry, 1, 0, 1, 1)
        grid.attach(self.install_from_setup_game_slug_checkbox, 0, 1, 1, 1)
        grid.attach(self.install_from_setup_game_slug_entry, 1, 1, 1, 1)

        self.install_from_setup_game_name_entry.connect("changed", self.on_install_from_setup_game_name_changed)
        self.install_from_setup_game_slug_checkbox.connect("toggled", self.on_install_from_setup_game_slug_toggled)

        explanation = _(
            "Enter the name of the game you will install.\n\nWhen you click 'Install' below, "
            "the installer window will appear and guide you through a simple installation.\n\n"
            "It will prompt you for a setup executable, and will use Wine to install it.\n\n"
            "If you know the Lutris identifier for the game, you can provide it for improved "
            "Lutris integration, such as Lutris provided banners."
        )

        grid.attach(self._get_explanation_label(explanation), 0, 2, 2, 1)

        preset_label = Gtk.Label(_("Installer preset:"), visible=True)
        grid.attach(preset_label, 0, 3, 1, 1)

        self.installer_presets.append(["win11", _("Windows 11 64-bit")])
        self.installer_presets.append(["win10", _("Windows 10 64-bit (Default)")])
        self.installer_presets.append(["win7", _("Windows 7 64-bit")])

        wine_version = LutrisConfig(runner_slug="wine").runner_config.get("version")
        if wine_version != GE_PROTON_LATEST and not is_proton_version(wine_version):
            self.installer_presets.append(["winxp", _("Windows XP 32-bit")])
            self.installer_presets.append(["win98", _("Windows 98 32-bit")])

        renderer_text = Gtk.CellRendererText()
        self.install_preset_dropdown.pack_start(renderer_text, True)
        self.install_preset_dropdown.add_attribute(renderer_text, "text", 1)
        self.install_preset_dropdown.set_id_column(0)
        self.install_preset_dropdown.set_active_id("win10")

        grid.attach(self.install_preset_dropdown, 1, 3, 1, 1)
        self.install_preset_dropdown.set_halign(Gtk.Align.START)

        locale_label = Gtk.Label(_("Locale:"), visible=True)
        locale_label.set_xalign(0)
        grid.attach(locale_label, 0, 4, 1, 1)

        locale_list = sysoptions.get_locale_choices()
        for locale_humanized, locale in locale_list:
            self.installer_locale.append([locale, _(locale_humanized)])

        locale_renderer_text = Gtk.CellRendererText()
        self.install_locale_dropdown.pack_start(locale_renderer_text, True)
        self.install_locale_dropdown.add_attribute(locale_renderer_text, "text", 1)
        self.install_locale_dropdown.set_id_column(0)
        self.install_locale_dropdown.set_active(0)

        grid.attach(self.install_locale_dropdown, 1, 4, 1, 1)
        self.install_locale_dropdown.set_halign(Gtk.Align.START)

        grid.set_vexpand(True)
        return grid

    def on_install_from_setup_game_slug_entry_focus_out(self, *args):
        slug = slugify(self.install_from_setup_game_slug_entry.get_text())
        self.install_from_setup_game_slug_entry.set_text(slug)

    def present_install_from_setup_page(self):
        self.set_page_title_markup(_("<b>Select setup file</b>"))
        self.stack.present_page("install_from_setup")
        self.display_continue_button(self._on_install_setup_continue, label=_("_Install"))

    def on_install_from_setup_game_slug_toggled(self, checkbutton):
        self.install_from_setup_game_slug_entry.set_sensitive(checkbutton.get_active())
        self.on_install_from_setup_game_name_changed()

    def on_install_from_setup_game_name_changed(self, *_args):
        if not self.install_from_setup_game_slug_checkbox.get_active():
            name = self.install_from_setup_game_name_entry.get_text()
            proposed_slug = slugify(name) if name else ""
            self.install_from_setup_game_slug_entry.set_text(proposed_slug)

    def _on_install_setup_continue(self, button):
        name = self.install_from_setup_game_name_entry.get_text().strip()

        if not name:
            ErrorDialog(_("You must provide a name for the game you are installing."), parent=self)
            return

        if self.install_from_setup_game_slug_checkbox.get_active():
            game_slug = slugify(self.install_from_setup_game_slug_entry.get_text())
        else:
            game_slug = slugify(name)

        installer_preset = self.installer_presets[self.install_preset_dropdown.get_active()][0]
        arch = "win32" if installer_preset.startswith(("win98", "winxp")) else "win64"
        win_ver = installer_preset.split("-")[0]
        if win_ver != "win10":
            win_ver_task = {"task": {"name": "winetricks", "app": win_ver, "arch": arch}}
        else:
            win_ver_task = None

        locale_selected = self.installer_locale[self.install_locale_dropdown.get_active()][0]

        installer = {
            "name": name,
            "version": _("Setup file"),
            "slug": game_slug + "-setup",
            "game_slug": game_slug,
            "runner": "wine",
            "script": {
                "game": {"exe": AUTO_WIN32_EXE, "prefix": "$GAMEDIR"},
                "files": [{"setupfile": "N/A:%s" % _("Select the setup file")}],
                "installer": [{"task": {"name": "wineexec", "executable": "setupfile", "arch": arch}}],
                "system": {"env": {"LC_ALL": locale_selected}},
            },
        }
        if win_ver_task:
            installer["script"]["installer"].insert(0, win_ver_task)
        application = Gio.Application.get_default()
        application.show_installer_window([installer])
        self.destroy()

    # Install from Script Page

    def install_from_script(self):
        """Install from a YAML file"""
        self.stack.navigate_to_page(self.present_install_from_script_page)

    def create_install_from_script_page(self):
        grid = Gtk.Grid(row_spacing=6, column_spacing=6)
        label = self._get_label(_("Script file"))
        grid.attach(label, 0, 0, 1, 1)
        grid.attach(self.install_script_file_chooser, 1, 0, 1, 1)
        self.install_script_file_chooser.set_hexpand(True)

        explanation = _(
            "Lutris install scripts are YAML files that guide Lutris through "
            "the installation process.\n\n"
            "They can be obtained on Lutris.net, or written by hand.\n\n"
            "When you click 'Install' below, the installer window will "
            "appear and load the script, and it will guide the process from there."
        )

        grid.attach(self._get_explanation_label(explanation), 0, 1, 2, 1)
        return grid

    def present_install_from_script_page(self):
        self.set_page_title_markup(_("<b>Select a Lutris installer</b>"))
        self.stack.present_page("install_from_script")
        self.display_continue_button(self.on_continue_install_from_script_clicked, label=_("_Install"))

    def on_continue_install_from_script_clicked(self, _widget):
        path = self.install_script_file_chooser.get_text()
        if not path:
            ErrorDialog(_("You must select a script file to install."), parent=self)
        elif not os.path.isfile(path):
            ErrorDialog(_("No file exists at '%s'.") % path, parent=self)
        else:
            installers = get_installers(installer_file=path)
            application = Gio.Application.get_default()
            application.show_installer_window(installers)
            self.destroy()

    # Install ROM Page

    def import_rom(self):
        """Install from a YAML file"""
        self.stack.navigate_to_page(self.present_import_rom_page)

    def create_import_rom_page(self):
        grid = Gtk.Grid(row_spacing=6, column_spacing=6)
        label = self._get_label(_("ROM file"))
        grid.attach(label, 0, 0, 1, 1)
        grid.attach(self.import_rom_file_chooser, 1, 0, 1, 1)
        self.import_rom_file_chooser.set_hexpand(True)

        explanation = _(
            "Lutris will identify ROMs via its MD5 hash and download game "
            "information from Lutris.net.\n\n"
            "The ROM data used for this comes from  TOSEC, No-Intro and Redump projects.\n\n"
            "When you click 'Install' below, the process of installing the games will "
            "begin."
        )

        grid.attach(self._get_explanation_label(explanation), 0, 1, 2, 1)
        return grid

    def present_import_rom_page(self):
        self.set_page_title_markup(_("<b>Select ROMs</b>"))
        self.stack.present_page("import_rom")
        self.display_continue_button(self.on_continue_import_rom_clicked, label=_("_Install"))

    def on_continue_import_rom_clicked(self, _widget):
        paths = []
        for path, _dirs, files in os.walk(self.import_rom_file_chooser.get_text()):
            for file in files:
                paths.append(os.path.join(path, file))

        if not paths:
            ErrorDialog(_("You must select ROMs to install."), parent=self)
        else:
            application = Gio.Application.get_default()
            dialog = ImportGameDialog(paths, parent=application.window)
            dialog.show()
            self.destroy()

    # Add Local Game

    def add_local_game(self):
        """Manually configure game"""
        # We use the LutrisWindow as the parent because we would
        # destroy this window before the AddGameDialog could disconnect.
        # We've tried to be clever here, but it didn't work reliably.
        # This does center the AddGameDialog over the main window, which
        # isn't terrible.
        application = Gio.Application.get_default()
        AddGameDialog(parent=application.window)
        self.destroy()

    # Subtitle Label

    def set_page_title_markup(self, markup):
        """Places some text at the top of the page; set markup to 'None' to remove it."""
        if markup:
            self.page_title_label.set_markup(markup)
            self.page_title_label.show()
        else:
            self.page_title_label.hide()

    # Continue Button

    def display_continue_button(self, handler, label=_("_Continue"), suggested_action=True):
        self.continue_button.set_label(label)
        style_context = self.continue_button.get_style_context()

        if suggested_action:
            style_context.add_class("suggested-action")
        else:
            style_context.remove_class("suggested-action")

        if self.continue_handler:
            self.continue_button.disconnect(self.continue_handler)

        self.continue_handler = self.continue_button.connect("clicked", handler)

        self.continue_button.show()
        self.cancel_button.set_label(_("Cancel"))
        self.stack.set_cancel_allowed(True)

    def display_cancel_button(self, label=_("Cancel")):
        self.cancel_button.set_label(label)
        self.stack.set_cancel_allowed(True)
        self.continue_button.hide()

    def display_no_continue_button(self):
        self.continue_button.hide()
        self.stack.set_cancel_allowed(False)

        if self.continue_handler:
            self.continue_button.disconnect(self.continue_handler)
            self.continue_handler = None

    # Implementation

    def _get_icon(self, name, small=False):
        if small:
            size = Gtk.IconSize.MENU
        else:
            size = Gtk.IconSize.DND
        icon = Gtk.Image.new_from_icon_name(name, size)
        icon.show()
        return icon

    def _get_label(self, text):
        label = Gtk.Label(visible=True)
        label.set_markup(text)
        label.set_alignment(0, 0.5)
        return label

    def _get_explanation_label(self, markup):
        label = Gtk.Label(visible=True, margin_right=12, margin_left=12, margin_top=12, margin_bottom=12)
        label.set_markup(markup)
        label.set_line_wrap(True)
        return label

    def _get_listbox_row(self, left_icon_name, text, subtext, right_icon_name=""):
        row = Gtk.ListBoxRow(visible=True)
        row.set_selectable(False)
        row.set_activatable(True)

        box = Gtk.Box(spacing=12, margin_right=12, margin_left=12, margin_top=12, margin_bottom=12, visible=True)

        if left_icon_name:
            icon = self._get_icon(left_icon_name)
            box.pack_start(icon, False, False, 0)
        label = self._get_label(f"<b>{text}</b>\n{subtext}")
        box.pack_start(label, True, True, 0)
        if left_icon_name:
            next_icon = self._get_icon(right_icon_name, small=True)
            box.pack_start(next_icon, False, False, 0)
        row.add(box)
        return row



import json
import os
import shutil

from lutris.settings import CACHE_DIR
from lutris.util import system
from lutris.util.files import get_folder_contents
from lutris.util.log import logger

FIRMWARE_CACHE_PATH = os.path.join(CACHE_DIR, "bios-files.json")


def scan_firmware_directory(target_directory: str):
    """Scans a target directory for firmwares and generates/updates the JSON 'firmware cache'
    file with relevant details and hashes for each file within the directory"""

    with open(FIRMWARE_CACHE_PATH, "w+") as firmwares_cache:
        json.dump(get_folder_contents(target_directory, with_hash=True), firmwares_cache, indent=2)


def get_firmware(target_firmware_name: str, target_firmware_checksum: str, runner_system_path: str):
    """Given a target firmware's name and checksum and the target runner's system directory, searches the
    user's BIOS cache for a firmware matching the checksum and places it under the provided system directory
    and name"""

    if not system.path_exists(FIRMWARE_CACHE_PATH):
        logger.error(f"Firmware {FIRMWARE_CACHE_PATH} not found.")
        return

    with open(FIRMWARE_CACHE_PATH) as bios_cache_data:
        bios_cache = json.load(bios_cache_data)

    for cached_firmware_record in bios_cache:
        # The checksum of the installed firmware we're looking at matches our target
        if cached_firmware_record.get("md5_hash") == target_firmware_checksum:
            system.create_folder(runner_system_path)
            shutil.copyfile(
                cached_firmware_record["name"],
                os.path.join(runner_system_path, target_firmware_name),
            )
            logger.info(f"Firmware {target_firmware_name} found and copied to {runner_system_path}")



# Standard Library
import binascii
import struct

# Lutris Modules
from lutris.util.gamecontrollerdb import GameControllerDB
from lutris.util.log import logger

try:
    import evdev
except ImportError:
    evdev = None
except AttributeError as err:
    # 'evdev' versions 1.5 and earlier are incompatible with Python 3.11
    # and produce this exception; we won't be able to use these.
    logger.exception("python3-evdev failed to load, and won't be available: %s", err)
    evdev = None


def get_devices():
    if not evdev:
        logger.warning("python3-evdev not installed, controller support not available")
        return []
    _devices = []
    for dev in evdev.list_devices():
        try:
            _devices.append(evdev.InputDevice(dev))
        except RuntimeError:
            pass
    return _devices


def get_joypads():
    """Return a list of tuples with the device and the joypad name"""
    return [(dev.path, dev.name) for dev in get_devices()]


def read_button(device):
    """Reference function for reading controller buttons and axis values.
    Not to be used as is.
    """
    # pylint: disable=no-member
    for event in device.read_loop():
        if event.type == evdev.ecodes.EV_KEY and event.value == 0:
            print("button %s (%s): %s" % (event.code, hex(event.code), event.value))
        if event.type == evdev.ecodes.EV_ABS:
            sticks = (0, 1, 3, 4)
            if event.code not in sticks or abs(event.value) > 5000:
                print("axis %s (%s): %s" % (event.code, hex(event.code), event.value))


def get_sdl_identifier(device_info):
    device_identifier = struct.pack(
        "<LLLL",
        device_info.bustype,
        device_info.vendor,
        device_info.product,
        device_info.version,
    )
    return binascii.hexlify(device_identifier).decode()


def get_controller_mappings():
    devices = get_devices()
    controller_db = GameControllerDB()

    controllers = []

    for device in devices:
        guid = get_sdl_identifier(device.info)
        if guid in controller_db.controllers:
            controllers.append((device, controller_db[guid]))

    return controllers



"""Runner loaders"""

__all__ = [
    "atari800",
    "azahar",
    "cemu",
    "dolphin",
    "dosbox",
    "duckstation",
    "easyrpg",
    "flatpak",
    "fsuae",
    "hatari",
    "jzintv",
    "libretro",
    "linux",
    "mame",
    "mednafen",
    "mupen64plus",
    "o2em",
    "osmose",
    "pcsx2",
    "pico8",
    "redream",
    "reicast",
    "rpcs3",
    "ryujinx",
    "scummvm",
    "shadps4",
    "snes9x",
    "steam",
    "vice",
    "vita3k",
    "web",
    "wine",
    "xemu",
    "xenia",
    "yuzu",
    "zdoom",
]

from lutris.exceptions import LutrisError, MisconfigurationError

ADDON_RUNNERS = {}
_cached_runner_human_names = {}


class InvalidRunnerError(MisconfigurationError):
    """Raise if a runner name is used that is not known to Lutris."""


class RunnerInstallationError(LutrisError):
    """Raised if the attempt to install a runner fails, perhaps because
    of invalid data from a server."""


class NonInstallableRunnerError(LutrisError):
    """Raised if installed a runner that Lutris cannot install, like Flatpak.
    These must be installed separately."""


def get_runner_module(runner_name):
    if not is_valid_runner_name(runner_name):
        raise InvalidRunnerError("Invalid runner name '%s'" % runner_name)
    module = __import__("lutris.runners.%s" % runner_name, globals(), locals(), [runner_name], 0)
    if not module:
        raise InvalidRunnerError("Runner module for '%s' could not be imported." % runner_name)
    return module


def import_runner(runner_name):
    """Dynamically import a runner class."""
    if runner_name in ADDON_RUNNERS:
        return ADDON_RUNNERS[runner_name]

    runner_module = get_runner_module(runner_name)
    return getattr(runner_module, runner_name)


def import_task(runner, task):
    """Return a runner task."""
    runner_module = get_runner_module(runner)
    return getattr(runner_module, task)


def get_installed(sort=True):
    """Return a list of installed runners (class instances)."""
    installed = []
    for runner_name in __all__:
        runner = import_runner(runner_name)()
        if runner.is_installed():
            installed.append(runner)
    return sorted(installed) if sort else installed


def inject_runners(runners):
    for runner_name in runners:
        if runner_name not in __all__:
            ADDON_RUNNERS[runner_name] = runners[runner_name]
            __all__.append(runner_name)
    _cached_runner_human_names.clear()


def get_runner_names():
    return __all__


def is_valid_runner_name(runner_name: str) -> bool:
    return runner_name in __all__


def get_runner_human_name(runner_name):
    """Returns a human-readable name for a runner; as a convenience, if the name
    is falsy (None or blank) this returns an empty string. Provides caching for the
    names."""
    if runner_name:
        if runner_name not in _cached_runner_human_names:
            try:
                _cached_runner_human_names[runner_name] = import_runner(runner_name)().human_name
            except InvalidRunnerError:
                _cached_runner_human_names[runner_name] = runner_name  # an obsolete runner
        return _cached_runner_human_names[runner_name]

    return ""



"""Utility functions for MAME"""

# Standard Library
import json
import os
from xml.etree import ElementTree

# Lutris Modules
from lutris import settings
from lutris.util.log import logger

CACHE_DIR = os.path.join(settings.CACHE_DIR, "mame")


def simplify_manufacturer(manufacturer):
    """Give simplified names for some manufacturers"""
    manufacturer_map = {
        "Amstrad plc": "Amstrad",
        "Apple Computer": "Apple",
        "Commodore Business Machines": "Commodore",
    }
    return manufacturer_map.get(manufacturer, manufacturer)


def is_game(machine):
    """Return True if the given machine game is an original arcade game
    Clones return False
    """
    return (
        machine.attrib["isbios"] == "no"
        and machine.attrib["isdevice"] == "no"
        and machine.attrib["runnable"] == "yes"
        and "romof" not in machine.attrib
        # FIXME: Filter by the machines that accept coins, but not like that
        # and "coin" in machine.find("input").attrib
    )


def has_software_list(machine):
    """Return True if the machine has an associated software list"""
    _has_software_list = False
    for elem in machine:
        if elem.tag == "device_ref" and elem.attrib["name"] == "software_list":
            _has_software_list = True
    return _has_software_list


def is_system(machine):
    """Given a machine XML tag, return True if it is a computer, console or
    handheld.
    """
    if (
        machine.attrib.get("runnable") == "no"
        or machine.attrib.get("isdevice") == "yes"
        or machine.attrib.get("isbios") == "yes"
    ):
        return False
    return has_software_list(machine)


def iter_machines(xml_path, filter_func=None):
    """Iterate through machine nodes in the MAME XML"""
    try:
        root = ElementTree.parse(xml_path).getroot()
    except Exception as ex:  # pylint: disable=broad-except
        logger.error("Failed to read MAME XML: %s", ex)
        return []
    for machine in root:
        if filter_func and not filter_func(machine):
            continue
        yield machine


def get_machine_info(machine):
    """Return human readable information about a machine node"""
    return {
        "description": machine.find("description").text,
        "manufacturer": simplify_manufacturer(machine.find("manufacturer").text),
        "year": machine.find("year").text,
        "roms": [rom.attrib for rom in machine.findall("rom")],
        "ports": [port.attrib for port in machine.findall("port")],
        "devices": [
            {
                "info": device.attrib,
                "name": "".join([instance.attrib["name"] for instance in device.findall("instance")]),
                "briefname": "".join([instance.attrib["briefname"] for instance in device.findall("instance")]),
                "extensions": [extension.attrib["name"] for extension in device.findall("extension")],
            }
            for device in machine.findall("device")
        ],
        "input": machine.find("input").attrib,
        "driver": machine.find("driver").attrib,
    }


def get_supported_systems(xml_path, force=False):
    """Return supported systems (computers and consoles) supported.
    From the full XML list extracted from MAME, filter the systems that are
    runnable, not clones and have the ability to run software.
    """
    systems_cache_path = os.path.join(CACHE_DIR, "systems.json")
    if os.path.exists(systems_cache_path) and not force:
        with open(systems_cache_path, "r", encoding="utf-8") as systems_cache_file:
            try:
                systems = json.load(systems_cache_file)
            except json.JSONDecodeError:
                logger.error("Failed to read systems cache %s", systems_cache_path)
                systems = None
        if systems:
            return systems
    systems = {machine.attrib["name"]: get_machine_info(machine) for machine in iter_machines(xml_path, is_system)}
    if not systems:
        return {}
    with open(systems_cache_path, "w", encoding="utf-8") as systems_cache_file:
        json.dump(systems, systems_cache_file, indent=2)
    return systems


def get_games(xml_path):
    """Return a list of all games"""
    return {machine.attrib["name"]: get_machine_info(machine) for machine in iter_machines(xml_path, is_game)}
